from tkinter import *
from tkinter import ttk
from tkinter import filedialog
from tkinter import font as tkfont
from typing import Dict, Union, Tuple, Iterable
from pathlib import Path, WindowsPath
from os import sep, utime
import time
import logging

import toml
import attr
from attr.validators import instance_of

from .config_parser import Config
import pysight

from pysight.nd_hist_generator.movie import ImagingSoftware


def is_positive(instance, attribute, value):
    if value < 0:
        return ValueError("TAG Bit value has to be greater than 0.")


def end_is_greater(instance, attribute, value):
    if value < instance.start:
        return ValueError("TAG Bit 'end' value has to be equal or greater to 'start'.")


@attr.s(slots=True)
class TagBits(object):
    """
    Storage for TAG bits
    """

    value = attr.ib(default="None", validator=instance_of(str))
    start = attr.ib(default=0, validator=[instance_of(int), is_positive])
    end = attr.ib(default=1, validator=[instance_of(int), is_positive, end_is_greater])


class GuiAppLst:
    """
    Main GUI for the multiscaler code.
    Note - class variables should contain "entry" in their name if they point
    to an entry TTK object. Also, no variable should contain "root" in its name.
    """

    def __init__(self):

        self.root = Tk()
        self._data_sources = (
            "PMT1",
            "PMT2",
            "PMT3",
            "PMT4",
            "Lines",
            "Frames",
            "Laser",
            "TAG Lens",
            "Empty",
        )
        self.root.title(f"PySight \uFF5C  PBLab \uFF5C v{pysight.__version__}")
        self.root.rowconfigure(16, weight=1)
        self.root.columnconfigure(16, weight=1)
        main_frame = ttk.Frame(self.root, width=1000, height=1300)
        main_frame.grid(column=0, row=0)
        main_frame["borderwidth"] = 2
        style = ttk.Style()
        style.theme_use("clam")
        self.normal_font = tkfont.Font(family="Helvetica", size=10)
        self.bold_font = tkfont.Font(family="Helvetica", size=12, weight="bold")
        self.config_row = 14
        self.__create_vars()

        # Run widgets
        self.__browse_file(main_frame)
        self.__advanced_win(main_frame)
        self.__input_channels(main_frame)
        self.__num_of_frames(main_frame)
        self.__outputs(main_frame)
        self.__image_size(main_frame)
        self.__tag_bits(main_frame)
        self.__imaging_software(main_frame)

        # Only saving\loading functions after this point
        self.__save_cfg(main_frame)
        self.__load_cfg(main_frame)
        self.__load_last_used_cfg(main_frame)

        # Define the last quit button and wrap up GUI
        quit_button = ttk.Button(main_frame, text="Start", command=self.root.destroy)
        quit_button.grid(row=16, column=2, sticky="ns")

        self.root.bind("<Return>", self.__dest)
        for child in main_frame.winfo_children():
            child.grid_configure(padx=3, pady=2)

        self.root.wait_window()

    def __dest(self, event):
        self.root.destroy()

    def __create_vars(self):
        self.debug = BooleanVar(value=False)
        self.phase = DoubleVar(value=-2.78)
        self.reprate = DoubleVar(
            value=80e6
        )  # 80e6 for the Chameleon, 0 to raise ZeroDivisionError
        self.gating = BooleanVar(
            value=False
        )  # difference between pulse and arrival to sample
        self.binwidth = DoubleVar(value=800e-12)
        self.tag_freq = DoubleVar(value=0.189e6)
        self.tag_pulses = IntVar(value=1)
        self.tag_offset = IntVar(value=0)
        self.fill_frac = DoubleVar(value=72.0)  # percent
        self.bidir = BooleanVar(value=False)
        self.keep_unidir = BooleanVar(value=False)
        self.flim: BooleanVar = BooleanVar(value=False)
        self.censor: BooleanVar = BooleanVar(value=False)
        self.line_freq = DoubleVar(value=7930.0)  # Hz
        self.sweeps_as_lines = BooleanVar(value=False)
        self.frame_delay = DoubleVar(value=0.001)  # sec
        self.interleaved = BooleanVar(value=False)

    def __browse_file(self, main_frame):
        file_row = 0
        self.filename = StringVar(value="")
        browse_button = ttk.Button(main_frame, text="Browse", command=self.__browsefunc)
        browse_button.grid(column=0, row=file_row, sticky="ns")

        browse_entry = ttk.Entry(main_frame, textvariable=self.filename, width=80)
        browse_entry.grid(column=1, row=file_row, sticky="we", columnspan=2)

    def __imaging_software(self, main_frame):
        imaging_software_label = ttk.Label(
            main_frame, text="Imaging Software", font=self.bold_font
        )
        imaging_software_label.grid(row=5, column=2, sticky="ns")
        self.imaging_software = StringVar()
        cb_image = ttk.Combobox(
            main_frame, textvariable=self.imaging_software, width=10
        )
        cb_image.grid(row=6, column=2, sticky="ns")
        cb_image.set(ImagingSoftware.SCANIMAGE.value)
        cb_image["values"] = [item.value for item in ImagingSoftware]

    def __input_channels(self, main_frame):
        # Comboboxes
        inputs_row = 1
        input_channels_label = ttk.Label(
            main_frame,
            text="Input Channels                         ",
            font=self.bold_font,
        )
        input_channels_label.grid(column=0, row=inputs_row, columnspan=2)
        self.input_start = StringVar()
        self.input_stop1 = StringVar()
        self.input_stop2 = StringVar()
        self.input_stop3 = StringVar()
        self.input_stop4 = StringVar()
        self.input_stop5 = StringVar()
        mb1 = ttk.Combobox(main_frame, textvariable=self.input_start, width=10)
        mb1.grid(column=1, row=inputs_row + 1, sticky="w")
        mb1.set("PMT1")
        mb1["values"] = self._data_sources
        mb2 = ttk.Combobox(main_frame, textvariable=self.input_stop1, width=10)
        mb2.grid(column=1, row=inputs_row + 2, sticky="w")
        mb2.set("Empty")
        mb2["values"] = self._data_sources
        mb3 = ttk.Combobox(main_frame, textvariable=self.input_stop2, width=10)
        mb3.grid(column=1, row=inputs_row + 3, sticky="w")
        mb3.set("Lines")
        mb3["values"] = self._data_sources
        mb4 = ttk.Combobox(main_frame, textvariable=self.input_stop3, width=10)
        mb4.grid(column=1, row=inputs_row + 4, sticky="w")
        mb4.set("Empty")
        mb4["values"] = self._data_sources
        mb5 = ttk.Combobox(main_frame, textvariable=self.input_stop4, width=10)
        mb5.grid(column=1, row=inputs_row + 5, sticky="w")
        mb5.set("Empty")
        mb5["values"] = self._data_sources
        mb6 = ttk.Combobox(main_frame, textvariable=self.input_stop5, width=10)
        mb6.grid(column=1, row=inputs_row + 6, sticky="w")
        mb6.set("Empty")
        mb6["values"] = self._data_sources

        # Labels
        input_channel_1 = ttk.Label(main_frame, text="START", font=self.normal_font)
        input_channel_1.grid(column=0, row=inputs_row + 1, sticky="ns")
        input_channel_2 = ttk.Label(main_frame, text="STOP1", font=self.normal_font)
        input_channel_2.grid(column=0, row=inputs_row + 2, sticky="ns")
        input_channel_3 = ttk.Label(main_frame, text="STOP2", font=self.normal_font)
        input_channel_3.grid(column=0, row=inputs_row + 3, sticky="ns")
        input_channel_4 = ttk.Label(main_frame, text="STOP3", font=self.normal_font)
        input_channel_4.grid(column=0, row=inputs_row + 4, sticky="ns")
        input_channel_5 = ttk.Label(main_frame, text="STOP4", font=self.normal_font)
        input_channel_5.grid(column=0, row=inputs_row + 5, sticky="ns")
        input_channel_6 = ttk.Label(main_frame, text="STOP5", font=self.normal_font)
        input_channel_6.grid(column=0, row=inputs_row + 6, sticky="ns")

    def __num_of_frames(self, main_frame):

        # Number of frames in the data
        frame_label = ttk.Label(
            main_frame, text="Number of frames", font=self.normal_font
        )
        frame_label.grid(column=2, row=4, sticky="w")

        self.num_of_frames = IntVar(value=1)
        self.num_frames_entry = ttk.Entry(
            main_frame, textvariable=self.num_of_frames, width=3
        )
        self.num_frames_entry.grid(column=2, row=4, sticky="ns")
        self.num_frames_entry.config(state="disabled")

        # Disable number of frames unless all inputs but one are empty
        self.input_start.trace("w", self.__check_if_empty)
        self.input_start.trace("w", self.__check_if_tag_lens_exists)
        self.input_stop1.trace("w", self.__check_if_empty)
        self.input_stop1.trace("w", self.__check_if_tag_lens_exists)
        self.input_stop2.trace("w", self.__check_if_empty)
        self.input_stop2.trace("w", self.__check_if_tag_lens_exists)
        self.input_stop3.trace("w", self.__check_if_empty)
        self.input_stop3.trace("w", self.__check_if_tag_lens_exists)
        self.input_stop4.trace("w", self.__check_if_empty)
        self.input_stop4.trace("w", self.__check_if_tag_lens_exists)
        self.input_stop5.trace("w", self.__check_if_empty)
        self.input_stop5.trace("w", self.__check_if_tag_lens_exists)

    def __outputs(self, main_frame):
        """ Wanted outputs """
        outputs_row = 9
        outputs_column = 2
        outputs_label = ttk.Label(main_frame, text="Outputs", font=self.bold_font)
        outputs_label.grid(column=outputs_column, row=outputs_row - 1, sticky="ns")

        self.summed = BooleanVar(value=False)
        summed_array = ttk.Checkbutton(
            main_frame, text="Summed Stack", variable=self.summed
        )
        summed_array.grid(column=outputs_column, row=outputs_row, sticky="ns")
        self.memory = BooleanVar(value=False)
        in_memory = ttk.Checkbutton(main_frame, text="In Memory", variable=self.memory)
        in_memory.grid(column=outputs_column, row=outputs_row + 1, sticky="ns")
        self.stack = BooleanVar(value=True)
        tif = ttk.Checkbutton(main_frame, text="Full Stack", variable=self.stack)
        tif.grid(column=outputs_column, row=outputs_row + 2, sticky="ns")

    def __image_size(self, main_frame):
        image_size_row = 1
        image_size_label = ttk.Label(main_frame, text="Image Size", font=self.bold_font)
        image_size_label.grid(column=2, row=image_size_row, sticky="ns", columnspan=1)
        x_size_label = ttk.Label(main_frame, text="X", font=self.normal_font)
        x_size_label.grid(column=2, row=image_size_row + 1, sticky="w")
        y_size_label = ttk.Label(main_frame, text="Y", font=self.normal_font)
        y_size_label.grid(column=2, row=image_size_row + 1, sticky="ns")
        z_size_label = ttk.Label(main_frame, text="Z", font=self.normal_font)
        z_size_label.grid(column=2, row=image_size_row + 1, sticky="e")

        self.x_pixels = IntVar(value=512)
        self.y_pixels = IntVar(value=512)
        self.z_pixels = IntVar(value=1)

        x_pixels_entry = ttk.Entry(main_frame, textvariable=self.x_pixels, width=5)
        x_pixels_entry.grid(column=2, row=image_size_row + 2, sticky="w")
        y_pixels_entry = ttk.Entry(main_frame, textvariable=self.y_pixels, width=5)
        y_pixels_entry.grid(column=2, row=image_size_row + 2, sticky="ns")
        self.z_pixels_entry = ttk.Entry(main_frame, textvariable=self.z_pixels, width=5)
        self.z_pixels_entry.grid(column=2, row=image_size_row + 2, sticky="e")
        self.z_pixels_entry.config(state="disabled")

    def __debug(self, main_frame):
        """ Read a smaller portion of data for debugging """
        debug_check = ttk.Checkbutton(main_frame, text="Debug?", variable=self.debug)
        debug_check.grid(column=2, row=8, sticky="ns")

    def __interleaved(self, main_frame):
        """ Unmix two data channel in the same PMT1 analog channel """
        inter_check = ttk.Checkbutton(
            main_frame, text="Interleaved?", variable=self.interleaved
        )
        inter_check.grid(column=2, row=7, sticky="ns")

    def __mirror_phase(self, main_frame):
        phase_text = ttk.Label(main_frame, text="Mirror phase [us]: ")
        phase_text.grid(column=0, row=1, sticky="w")
        phase_entry = ttk.Entry(main_frame, textvariable=self.phase, width=8)
        phase_entry.grid(column=0, row=1, sticky="e")

    def __reprate(self, main_frame):
        """ Laser repetition rate"""

        laser1_label = ttk.Label(main_frame, text="Laser nominal rep. rate (FLIM) [Hz]")
        laser1_label.grid(column=2, row=6, sticky="ns")
        reprate_entry = ttk.Entry(main_frame, textvariable=self.reprate, width=11)
        reprate_entry.grid(column=3, row=6, sticky="ns")

    def __gating(self, main_frame):
        self.gating_check = ttk.Checkbutton(
            main_frame, text="With Gating?", variable=self.gating
        )
        self.gating_check.grid(column=2, row=5, sticky="ns")
        self.gating_check.config(state="disabled")

    def __binwidth(self, main_frame):
        """ Binwidth of Multiscaler (for FLIM) """

        binwidth_label = ttk.Label(main_frame, text="Multiscaler binwidth [sec]")
        binwidth_label.grid(column=2, row=1, sticky="ns")
        binwidth_entry = ttk.Entry(main_frame, textvariable=self.binwidth, width=9)
        binwidth_entry.grid(column=3, row=1, sticky="ns")

    def __tag_lens(self, main_frame):
        """ TAG lens nominal frequency """
        tag_row = 7
        tag_label = ttk.Label(
            main_frame,
            text="     TAG nominal freq. [Hz]\noffset [deg]                   n. pulses",
        )
        tag_label.grid(column=0, row=tag_row, columnspan=2, sticky="w")
        tag_label_entry = ttk.Entry(main_frame, textvariable=self.tag_freq, width=10)
        tag_label_entry.grid(column=0, row=tag_row + 1, sticky="ns")

        tag_pulses_entry = ttk.Entry(main_frame, textvariable=self.tag_pulses, width=3)
        tag_pulses_entry.grid(column=0, row=tag_row + 1, sticky="e")
        tag_pulses_entry.config(state="disabled")

        self.tag_offset_entry = ttk.Entry(
            main_frame, textvariable=self.tag_offset, width=3
        )
        self.tag_offset_entry.grid(column=0, row=tag_row + 1, sticky="w")

    def __tag_bits(self, main_frame):
        """ TAG bits """
        tag_bits_row = 9
        tag_bits_label = ttk.Label(
            main_frame, text="TAG Bits Allocation", font=self.bold_font
        )
        tag_bits_label.grid(column=1, row=tag_bits_row, sticky="ns")

        self.tag_bits = BooleanVar(value=False)
        tag_bit_check = ttk.Checkbutton(main_frame, text="Use?", variable=self.tag_bits)
        tag_bit_check.grid(column=1, row=tag_bits_row, sticky="w")

        self.bits_grp_1_start = IntVar(value=1)
        self.bits_grp_1_end = IntVar(value=3)
        self.bits_grp_2_start = IntVar(value=4)
        self.bits_grp_2_end = IntVar(value=5)
        self.bits_grp_3_start = IntVar(value=6)
        self.bits_grp_3_end = IntVar(value=16)

        self.bits_grp_1_label = StringVar()
        self.bits_grp_2_label = StringVar()
        self.bits_grp_3_label = StringVar()

        self.tag_bits_group_options = (
            "Power",
            "Slow axis",
            "Fast axis",
            "Z axis",
            "None",
        )

        bits_grp_1 = ttk.Combobox(
            main_frame, textvariable=self.bits_grp_1_label, width=10
        )
        bits_grp_1.grid(column=0, row=tag_bits_row + 1, sticky="e")
        bits_grp_1.set("None")
        bits_grp_1["values"] = self.tag_bits_group_options

        bits_grp_2 = ttk.Combobox(
            main_frame, textvariable=self.bits_grp_2_label, width=10
        )
        bits_grp_2.grid(column=0, row=tag_bits_row + 2, sticky="e")
        bits_grp_2.set("None")
        bits_grp_2["values"] = self.tag_bits_group_options

        bits_grp_3 = ttk.Combobox(
            main_frame, textvariable=self.bits_grp_3_label, width=10
        )
        bits_grp_3.grid(column=0, row=tag_bits_row + 3, sticky="e")
        bits_grp_3.set("None")
        bits_grp_3["values"] = self.tag_bits_group_options

        bits_grp_1_start_lab = ttk.Label(main_frame, text="\tStart")
        bits_grp_1_start_lab.grid(column=1, row=tag_bits_row + 1, sticky="w")
        bits_grp_1_start_ent = ttk.Entry(
            main_frame, textvariable=self.bits_grp_1_start, width=3
        )
        bits_grp_1_start_ent.grid(column=1, row=tag_bits_row + 1, sticky="ns")
        bits_grp_1_end_lab = ttk.Label(main_frame, text="End")
        bits_grp_1_end_lab.grid(column=1, row=tag_bits_row + 1, sticky="e")
        bits_grp_1_end_ent = ttk.Entry(
            main_frame, textvariable=self.bits_grp_1_end, width=3
        )
        bits_grp_1_end_ent.grid(column=2, row=tag_bits_row + 1, sticky="w")

        bits_grp_2_start_lab = ttk.Label(main_frame, text="\tStart")
        bits_grp_2_start_lab.grid(column=1, row=tag_bits_row + 2, sticky="w")
        bits_grp_2_start_ent = ttk.Entry(
            main_frame, textvariable=self.bits_grp_2_start, width=3
        )
        bits_grp_2_start_ent.grid(column=1, row=tag_bits_row + 2, sticky="ns")
        bits_grp_2_end_lab = ttk.Label(main_frame, text="End")
        bits_grp_2_end_lab.grid(column=1, row=tag_bits_row + 2, sticky="e")
        bits_grp_2_end_ent = ttk.Entry(
            main_frame, textvariable=self.bits_grp_2_end, width=3
        )
        bits_grp_2_end_ent.grid(column=2, row=tag_bits_row + 2, sticky="w")

        bits_grp_3_start_lab = ttk.Label(main_frame, text="\tStart")
        bits_grp_3_start_lab.grid(column=1, row=tag_bits_row + 3, sticky="w")
        bits_grp_3_start_ent = ttk.Entry(
            main_frame, textvariable=self.bits_grp_3_start, width=3
        )
        bits_grp_3_start_ent.grid(column=1, row=tag_bits_row + 3, sticky="ns")
        bits_grp_3_end_lab = ttk.Label(main_frame, text="End")
        bits_grp_3_end_lab.grid(column=1, row=tag_bits_row + 3, sticky="e")
        bits_grp_3_end_ent = ttk.Entry(
            main_frame, textvariable=self.bits_grp_3_end, width=3
        )
        bits_grp_3_end_ent.grid(column=2, row=tag_bits_row + 3, sticky="w")

        self.tag_bits_dict = {}
        self.tag_bits_dict = {
            0: TagBits(
                value=self.bits_grp_1_label.get(),
                start=self.bits_grp_1_start.get(),
                end=self.bits_grp_1_end.get(),
            ),
            1: TagBits(
                value=self.bits_grp_2_label.get(),
                start=self.bits_grp_2_start.get(),
                end=self.bits_grp_2_end.get(),
            ),
            2: TagBits(
                value=self.bits_grp_3_label.get(),
                start=self.bits_grp_3_start.get(),
                end=self.bits_grp_3_end.get(),
            ),
        }

    def __fill_frac(self, main_frame):
        """ Percentage of time mirrors spend "inside" the image """

        fill_frac_text = ttk.Label(main_frame, text="Fill fraction [%]: ")
        fill_frac_text.grid(column=0, row=4, sticky="w")
        fill_frac_entry = ttk.Entry(main_frame, textvariable=self.fill_frac, width=8)
        fill_frac_entry.grid(column=0, row=4, sticky="e")

    def __browsefunc(self):
        filetypes = [("List files", "*.lst"), ("Pickle files", "*.p")]
        if self.filename.get() != "":
            self.filename.set(
                filedialog.askopenfilename(
                    filetypes=filetypes,
                    title="Choose a list or pickle file",
                    initialdir=str(Path(self.filename.get()).parent),
                )
            )
        else:
            self.filename.set(
                filedialog.askopenfilename(
                    filetypes=filetypes,
                    title="Choose a list or pickle file",
                    initialdir=".",
                )
            )

    def __check_if_empty(self, *args):
        list_of_values = [
            self.input_start.get(),
            self.input_stop1.get(),
            self.input_stop2.get(),
            self.input_stop3.get(),
            self.input_stop4.get(),
            self.input_stop5.get(),
        ]
        if 2 == list_of_values.count("Empty"):
            if "PMT1" in list_of_values or "PMT2" in list_of_values:
                self.num_frames_entry.config(state="normal")
            else:
                self.num_frames_entry.config(state="disabled")

    def __check_if_tag_lens_exists(self, *args):
        list_of_values = [
            self.input_start.get(),
            self.input_stop1.get(),
            self.input_stop2.get(),
            self.input_stop3.get(),
            self.input_stop4.get(),
            self.input_stop5.get(),
        ]
        if "TAG Lens" in list_of_values:
            self.z_pixels_entry.config(state="normal")
            # self.tag_offset_entry.config(state='normal')
        else:
            self.z_pixels_entry.config(state="disabled")
            # self.tag_offset_entry.config(state='disabled')

    def __bidir(self, main_frame):
        """ Checkbox for bi-directional scan """

        bidir_check = ttk.Checkbutton(
            main_frame, text="Bi-directional scan", variable=self.bidir
        )
        bidir_check.grid(column=0, row=5, sticky="ns")
        self.bidir.trace("w", self.__check_if_bidir)

    def __check_if_bidir(self, *args):
        if self.bidir:
            self.keep_unidir_check.config(state="normal")
        if not self.bidir:
            self.keep_unidir_check.config(state="disabled")

    def __keep_unidir_events(self, main_frame):
        """ Checkbox to see if events taken in the returning phase of a resonant mirror should be kept. """
        self.keep_unidir_check = ttk.Checkbutton(
            main_frame, text="Keep unidirectional?", variable=self.keep_unidir
        )
        self.keep_unidir_check.grid(column=0, row=6, sticky="ns")
        self.keep_unidir_check.config(state="disabled")

    def __flim(self, main_frame):
        """
        Defines the mapping between one pulse and the missing pulses.
        For example, downsampling factor of 8 means that every pulse that is
        received starts an event of 8 pulses, with the next recorded pulse being the 9th.
        :param main_frame: ttk.Frame
        """
        flim_check: ttk.Checkbutton = ttk.Checkbutton(
            main_frame, variable=self.flim, text="FLIM?"
        )
        flim_check.grid(row=2, column=2, sticky="ns")
        self.flim.trace("w", self.__check_if_flim)

    def __censor(self, main_frame):
        """
        If FLIM is active, this checkbox enables the use of censor correction on the generated images.
        :param main_frame: ttk.Frame
        """
        self.censor_check: ttk.Checkbutton = ttk.Checkbutton(
            main_frame, variable=self.censor, text="Censor Correction"
        )
        self.censor_check.grid(row=3, column=2, sticky="ns")
        self.censor_check.config(state="disabled")

    def __check_if_flim(self, *args):
        if self.flim:
            self.censor_check.config(state="normal")
            self.gating_check.config(state="normal")
        else:
            self.censor_check.config(state="disabled")
            self.gating_check.config(state="disabled")
        self.root.update_idletasks()

    def __line_freq(self, main_frame):
        """ Frequency of the line scanning mirror """
        line_freq_label = ttk.Label(main_frame, text="Line freq [Hz]: ")
        line_freq_label.grid(row=3, column=0, sticky="w")
        line_freq_entry = ttk.Entry(main_frame, textvariable=self.line_freq, width=8)
        line_freq_entry.grid(row=3, column=0, sticky="e")

    def __sweeps_as_lines(self, main_frame):
        """ Use the sweeps as lines for the image generation """
        sweeps_cb = ttk.Checkbutton(
            main_frame, variable=self.sweeps_as_lines, text="Sweeps as lines?"
        )
        sweeps_cb.grid(row=4, column=2, sticky="ns")

    def __advanced_win(self, main_frame):
        advanced_but = ttk.Button(
            main_frame, text="Advanced", command=self.__open_advanced
        )
        advanced_but.grid(row=13, column=2, sticky="ns")

    def __open_advanced(self, *args):
        self.advanced_win = Toplevel(self.root)
        frame = ttk.Frame(self.advanced_win, width=300, height=300)
        frame.grid(column=0, row=0)
        frame["borderwidth"] = 2
        style = ttk.Style()
        style.theme_use("clam")
        self.__setup_advanced_frame(frame)
        self.__gating(frame)
        self.__flim(frame)
        self.__censor(frame)
        self.__sweeps_as_lines(frame)
        self.__debug(frame)
        self.__mirror_phase(frame)
        self.__fill_frac(frame)
        self.__reprate(frame)
        self.__binwidth(frame)
        self.__keep_unidir_events(frame)
        self.__bidir(frame)
        self.__check_if_bidir(frame)
        self.__tag_lens(frame)
        self.__frame_delay(frame)
        self.__line_freq(frame)
        self.__interleaved(frame)
        for child in frame.winfo_children():
            child.grid_configure(padx=3, pady=2)

    def __setup_advanced_frame(self, frame):
        scan_lab = ttk.Label(frame, text="       Scanner Settings", font=self.bold_font)
        scan_lab.grid(row=0, column=0, sticky="ns")

        hardware_lab = ttk.Label(
            frame, text="               Hardware Settings", font=self.bold_font
        )
        hardware_lab.grid(row=0, column=2, sticky="ns")

    def __frame_delay(self, main_frame):
        frame_delay_label = ttk.Label(main_frame, text="Frame delay [sec]: ")
        frame_delay_label.grid(row=2, column=0, sticky="w")
        frame_delay_entry = ttk.Entry(
            main_frame, textvariable=self.frame_delay, width=8
        )
        frame_delay_entry.grid(row=2, column=0, sticky="e")

    ####### ONLY SAVE\LOAD FUNCS AFTER THIS POINT ########

    def __save_cfg(self, main_frame):
        """ A button to write a .toml with current configs """
        config_label = ttk.Label(
            main_frame, text="Configuration File", font=self.bold_font
        )
        config_label.grid(column=1, row=self.config_row, sticky="ns")
        self.save_as: StringVar = StringVar(value="default")
        save_label = ttk.Label(main_frame, text="Config file name to save:")
        save_label.grid(
            column=0, row=self.config_row + 1, sticky="ns", columnspan=2, padx=10
        )
        save_entry = ttk.Entry(main_frame, textvariable=self.save_as, width=8)
        save_entry.grid(column=1, row=self.config_row + 1, sticky="e")
        save_button = ttk.Button(
            main_frame, text="Save cfg", command=self.__callback_save_cur_cfg
        )
        save_button.grid(column=1, row=self.config_row + 2, sticky="w")

    def __callback_save_cur_cfg(self) -> None:
        """
        Takes a GUIApp() instance and saves it to a .toml file
        """
        cfg_dict_to_save = Config.from_gui(self)
        cfg_dict_to_save.to_disk()

    def __load_cfg(self, main_frame: ttk.Frame):
        """
        Load a specific .toml file and change all variables accordingly
        """
        self.cfg_filename: StringVar = StringVar(value="default")
        load_button: Button = ttk.Button(
            main_frame, text="Load cfg", command=self.__browsecfg
        )
        load_button.grid(column=1, row=self.config_row + 2, sticky="e")

    def __browsecfg(self, new_cfg=None):
        if not new_cfg:
            self.cfg_filename.set(
                filedialog.askopenfilename(
                    filetypes=[("Config files", "*.toml")],
                    title=f"Choose a configuration file",
                    initialdir=str(Path(__file__).parents[1] / "configs"),
                )
            )
        else:
            self.cfg_filename.set(new_cfg)
        with open(self.cfg_filename.get(), "r") as f:
            self.config = toml.load(f)
            try:
                utime(self.cfg_filename.get(), (time.time(), time.time()))
            except PermissionError:
                pass
        self.__modify_vars()

    def __modify_vars(self):
        """
        With the dictionary loaded from the TOML file, change all variables
        """
        from_cfg_to_vars = self._build_config_dict()
        for cfg_key, cfg_val in self.config.items():
            if isinstance(cfg_val, dict):
                for inner_key, inner_val in cfg_val.items():
                    if isinstance(inner_val, dict):
                        for innner_key, innner_val in inner_val.items():
                            from_cfg_to_vars[innner_key].set(innner_val)
                    else:
                        from_cfg_to_vars[inner_key].set(inner_val)
            else:
                from_cfg_to_vars[cfg_key].set(cfg_val)
        self.root.update_idletasks()

    def __load_last_used_cfg(self, main_frame):
        direc = Path(__file__).parents[1] / "configs"
        all_cfg_files: Iterable = direc.glob("*.toml")
        latest_filename: str = ""
        latest_file_date: int = 0
        for cfg_file in all_cfg_files:
            cur_date_modified = cfg_file.stat()[8]
            if cur_date_modified > latest_file_date:
                latest_filename = str(cfg_file)
                latest_file_date = cur_date_modified

        if latest_filename != "":
            with open(latest_filename, "r") as f:
                try:
                    self.config = toml.load(f)
                except ValueError:
                    self.config = {}
            self.__modify_vars()

    def _build_config_dict(self):
        """ Helper method to populate a new GUI instance from a config file """
        from_config_to_vars = {
            "cfg_title": self.save_as,
            "stop1": self.input_stop1,
            "stop2": self.input_stop2,
            "stop3": self.input_stop3,
            "stop4": self.input_stop4,
            "stop5": self.input_stop5,
            "start": self.input_start,
            "num_of_frames": self.num_of_frames,
            "x_pixels": self.x_pixels,
            "y_pixels": self.y_pixels,
            "z_pixels": self.z_pixels,
            "imaging_software": self.imaging_software,
            "data_filename": self.filename,
            "summed": self.summed,
            "memory": self.memory,
            "stack": self.stack,
            "debug": self.debug,
            "phase": self.phase,
            "reprate": self.reprate,
            "gating": self.gating,
            "binwidth": self.binwidth,
            "tag_freq": self.tag_freq,
            "tag_pulses": self.tag_pulses,
            "tag_offset": self.tag_offset,
            "fill_frac": self.fill_frac,
            "bidir": self.bidir,
            "keep_unidir": self.keep_unidir,
            "flim": self.flim,
            "censor": self.censor,
            "line_freq": self.line_freq,
            "sweeps_as_lines": self.sweeps_as_lines,
            "frame_delay": self.frame_delay,
            "interleaved": self.interleaved,
            "tag_bits": self.tag_bits,
            "label1": self.bits_grp_1_label,
            "start1": self.bits_grp_1_start,
            "end1": self.bits_grp_1_end,
            "label2": self.bits_grp_2_label,
            "start2": self.bits_grp_2_start,
            "end2": self.bits_grp_2_end,
            "label3": self.bits_grp_3_label,
            "start3": self.bits_grp_3_start,
            "end3": self.bits_grp_3_end,
        }
        return from_config_to_vars


if __name__ == "__main__":
    app = GuiAppLst()



"""
__author__ = Hagai Har-Gil
"""
from unittest import TestCase
import pathlib

import zarr
import toml
import numpy as np

from pysight.main import main_data_readout


class TestEntirePipeline(TestCase):
    """ Integration tests for the entire application """

    def test_standard_pipeline(self):
        cfg_file = str(next(pathlib.Path(".").rglob("*jul.toml")).absolute())
        data_file = str(next(pathlib.Path(".").rglob("*jul.zarr")).absolute())
        with open(cfg_file, "r") as f:
            config: Dict[str, Any] = toml.load(f)
        out = main_data_readout(config)
        z = zarr.open(data_file, "r")
        np.testing.assert_array_equal(
            out.ch1.time_summed, np.array(z["/Summed Stack/Channel 1"])
        )

    def test_mscan_pipeline(self):
        cfg_file = str(next(pathlib.Path(".").rglob("*fly.toml")).absolute())
        data_file = str(next(pathlib.Path(".").rglob("*fly.zarr")).absolute())
        with open(cfg_file, "r") as f:
            config: Dict[str, Any] = toml.load(f)
        out = main_data_readout(config)
        z = zarr.open(data_file, "r")
        np.testing.assert_array_equal(
            out.ch1.time_summed, np.array(z["/Summed Stack/Channel 1"])
        )

    def test_tag_pipeline(self):
        cfg_file = str(next(pathlib.Path(".").rglob("*tag.toml")).absolute())
        data_file = str(next(pathlib.Path(".").rglob("*tag.zarr")).absolute())
        with open(cfg_file, "r") as f:
            config: Dict[str, Any] = toml.load(f)
        out = main_data_readout(config)
        z = zarr.open(data_file, "r")
        np.testing.assert_array_equal(
            out.ch1.time_summed, np.array(z["/Summed Stack/Channel 1"])
        )



from collections import namedtuple
from pprint import pprint

import pytest

from pysight.nd_hist_generator.frame_chunk import *
from pysight.nd_hist_generator.movie import *


def gen_data_df(frame_num=10, line_num=100, end=1000, channels=2):
    """
    Mock data for tests.
    Returns:
        df - The full DataFrame
        frames only
        lines only
        x pixels
        y pixels
    """
    photons = np.arange(0, end, dtype=np.uint64)
    channel = np.ones_like(photons)
    if channels == 2:
        channel[len(channel) // 2 :] = 2
    lines = np.linspace(0, end, num=line_num, endpoint=False, dtype=np.uint64)
    x_pix = int(len(photons) / len(lines))
    ones_lines = np.ones((1, int(len(photons) / len(lines))), dtype=np.uint64)
    frames = np.linspace(0, end, num=frame_num, dtype=np.uint64, endpoint=False)
    frames_ser = pd.Series(frames, index=frames)
    ones_frames = np.ones((1, int(len(photons) / len(frames))), dtype=np.uint64)
    lines = (np.atleast_2d(lines).T @ ones_lines).ravel()
    frames = (np.atleast_2d(frames).T @ ones_frames).ravel()
    assert len(lines) == len(frames) == len(photons)

    df = pd.DataFrame(
        {
            "abs_time": photons,
            "time_rel_line": photons - lines,
            "Lines": lines,
            "Frames": frames,
            "Channel": channel,
        }
    )
    df.set_index(["Channel", "Frames", "Lines"], drop=True, inplace=True)
    y_pix = x_pix
    lines_to_return = pd.Series(
        np.unique(lines), index=np.repeat(frames_ser, line_num // frame_num)
    )

    return df, frames_ser, lines_to_return, x_pix, y_pix


class TestFrameChunk:
    df, frames, lines, x, y = gen_data_df()
    movie_single = Movie(
        df,
        lines,
        outputs={"memory": True},
        line_delta=int(lines.diff().mean()),
        fill_frac=100.0,
        bidir=True,
        data_shape=(len(frames), x, y),
        frame_slices=(slice(frame) for frame in frames),
        frames=frames,
        frames_per_chunk=10,
    )
    df_dict = {
        1: df.xs(key=(1, 100), level=("Channel", "Frames"), drop_level=False),
        2: df.xs(key=(2, 100), level=("Channel", "Frames"), drop_level=False),
    }
    chunk_single = FrameChunk(
        movie=movie_single,
        df_dict=df_dict,
        frames_per_chunk=10,
        frames=frames,
        lines=lines,
    )

    movie_multi = Movie(
        df,
        lines=lines,
        outputs={"memory": True},
        line_delta=int(lines.diff().mean()),
        fill_frac=100.0,
        bidir=True,
        data_shape=(len(frames), x, y),
        frames=frames,
        frame_slices=(slice(frame) for frame in frames),
        frames_per_chunk=4,
    )
    sl = pd.IndexSlice[slice(1), slice(100, 400)]
    chunk_multi = FrameChunk(
        movie=movie_multi,
        df_dict={1: df.loc[sl, :]},
        frames_per_chunk=4,
        frames=frames,
        lines=lines,
    )

    def test_frame_edges_single_chunk(self):
        fr = self.chunk_single._FrameChunk__create_frame_edges()
        np.testing.assert_equal(
            fr,
            np.array(
                [0, 100, 200, 300, 400, 500, 600, 700, 800, 900, 901], dtype=np.uint64
            ),
        )

    def test_frame_edges_multiple_frames(self):
        sl = pd.IndexSlice[slice(1), slice(100, 400)]
        movie = Movie(
            self.df,
            self.lines,
            outputs={"memory": True},
            line_delta=int(self.lines.diff().mean()),
            fill_frac=100.0,
            bidir=True,
            data_shape=(len(self.frames), self.x, self.y),
            frames=self.frames,
            frame_slices=(slice(frame) for frame in self.frames),
            frames_per_chunk=4,
        )
        chunk_multi = FrameChunk(
            movie=movie,
            df_dict={1: self.df.loc[sl, :]},
            frames=pd.Series([100, 200, 300, 400], dtype=np.uint64),
            lines=pd.Series([10]),
            frames_per_chunk=4,
        )
        fr = chunk_multi._FrameChunk__create_frame_edges()
        np.testing.assert_equal(fr, np.array([100, 200, 300, 400, 401]))

    def test_line_edges_single_chunk(self):
        li = self.chunk_single._FrameChunk__create_line_edges()
        lines = np.arange(0, 1010, 10)
        np.testing.assert_equal(li, lines)

    def test_line_edges_multi_chunk(self):
        sl = pd.IndexSlice[slice(1), slice(100, 400)]
        movie = Movie(
            self.df,
            self.lines,
            outputs={"memory": True},
            line_delta=int(self.lines.diff().mean()),
            fill_frac=100.0,
            bidir=True,
            data_shape=(len(self.frames), self.x, self.y),
            frames=self.frames,
            frame_slices=(slice(frame) for frame in self.frames),
            frames_per_chunk=4,
        )
        chunk_multi = FrameChunk(
            movie=movie,
            df_dict={1: self.df.loc[sl, :]},
            frames=self.frames,
            lines=self.lines.loc[slice(100, 400)],
            frames_per_chunk=4,
        )
        li = chunk_multi._FrameChunk__create_line_edges()
        lines = np.arange(100, 510, 10)
        np.testing.assert_equal(li, lines)

    def test_col_edges_single_frame(self):
        cr = self.chunk_single._FrameChunk__create_col_edges()
        cols = np.arange(11)
        np.testing.assert_equal(cr, cols)

    def test_col_edges_multi_frame(self):
        cr = self.chunk_multi._FrameChunk__create_col_edges()
        cols = np.arange(11)
        np.testing.assert_equal(cr, cols)

    def test_linspace_along_sine_1_pix_z(self):
        sine = self.chunk_single._FrameChunk__linspace_along_sine()
        np.testing.assert_almost_equal(
            np.array([-0.999_995_03, 0.999_995_01]), sine.tolist()
        )

    def test_linspace_along_sine_100_pix_z(self):
        movie_for_sine = Movie(
            self.df,
            self.lines,
            data_shape=(1, 512, 512, 100),
            frame_slices=(1,),
            frames=self.frames,
        )
        chunk = FrameChunk(
            movie_for_sine,
            self.df_dict,
            frames_per_chunk=1,
            frames=self.frames,
            lines=self.lines,
        )
        sin = chunk._FrameChunk__linspace_along_sine()
        true = np.array(
            [
                -9.999_950_30e-01,
                -9.800_004_36e-01,
                -9.600_004_08e-01,
                -9.400_011_49e-01,
                -9.200_012_49e-01,
                -9.000_018_13e-01,
                -8.800_007_00e-01,
                -8.600_020_42e-01,
                -8.400_002_45e-01,
                -8.200_015_19e-01,
                -8.000_050_73e-01,
                -7.800_049_16e-01,
                -7.600_003_63e-01,
                -7.400_022_30e-01,
                -7.200_009_40e-01,
                -7.000_044_11e-01,
                -6.800_007_61e-01,
                -6.600_036_90e-01,
                -6.400_041_55e-01,
                -6.200_007_61e-01,
                -6.000_020_52e-01,
                -5.800_040_59e-01,
                -5.600_065_23e-01,
                -5.400_055_25e-01,
                -5.200_023_25e-01,
                -5.000_042_41e-01,
                -4.799_990_88e-01,
                -4.599_986_46e-01,
                -4.400_044_90e-01,
                -4.199_985_05e-01,
                -3.999_970_92e-01,
                -3.799_975_74e-01,
                -3.599_960_63e-01,
                -3.399_969_68e-01,
                -3.200_039_66e-01,
                -3.000_009_72e-01,
                -2.799_995_81e-01,
                -2.600_014_20e-01,
                -2.399_978_86e-01,
                -2.199_992_22e-01,
                -2.000_056_36e-01,
                -1.800_071_04e-01,
                -1.600_029_81e-01,
                -1.400_022_36e-01,
                -1.200_037_63e-01,
                -1.000_062_21e-01,
                -8.000_806_72e-02,
                -6.000_760_10e-02,
                -4.000_299_65e-02,
                -2.000_233_92e-02,
                -3.673_215_39e-06,
                1.999_499_42e-02,
                3.999_565_59e-02,
                6.000_026_79e-02,
                8.000_074_44e-02,
                9.999_891_11e-02,
                1.199_964_70e-01,
                1.399_949_62e-01,
                1.599_957_30e-01,
                1.799_998_78e-01,
                1.999_984_38e-01,
                2.200_018_11e-01,
                2.400_004_62e-01,
                2.599_943_26e-01,
                2.800_021_29e-01,
                2.999_939_64e-01,
                3.199_970_06e-01,
                3.399_994_63e-01,
                3.599_985_38e-01,
                3.800_000_28e-01,
                3.999_995_24e-01,
                4.200_009_14e-01,
                4.399_978_93e-01,
                4.600_010_02e-01,
                4.800_014_16e-01,
                4.999_978_79e-01,
                5.199_960_50e-01,
                5.399_993_41e-01,
                5.600_004_36e-01,
                5.799_980_74e-01,
                5.999_961_74e-01,
                6.199_949_97e-01,
                6.399_985_10e-01,
                6.599_981_71e-01,
                6.799_953_75e-01,
                6.999_991_65e-01,
                7.200_027_81e-01,
                7.399_972_88e-01,
                7.599_955_88e-01,
                7.800_003_19e-01,
                7.999_946_65e-01,
                8.199_973_14e-01,
                8.400_016_84e-01,
                8.599_982_93e-01,
                8.799_972_10e-01,
                8.999_986_10e-01,
                9.199_983_70e-01,
                9.399_986_43e-01,
                9.599_983_51e-01,
                9.799_989_74e-01,
                9.999_950_07e-01,
            ]
        )
        np.testing.assert_almost_equal(sin, true)


class TestMyHist:
    """Tests for my own implementation of an histogram."""

    def test_basic_indices(self):
        data = [np.array([5, 15, 25])]
        bins = [np.array([0, 10, 20, 30])]
        hist = HistWithIndex(data, bins)
        idx, _ = hist._get_indices_for_photons()
        np.testing.assert_equal(idx, np.array([1, 2, 3]))

    def test_out_of_bounds(self):
        data = [np.array([-1, 5, 30, 40])]
        bins = [np.array([0, 10, 20, 30])]
        hist = HistWithIndex(data, bins)
        idx, _ = hist._get_indices_for_photons()
        np.testing.assert_equal(idx, np.array([0, 1, 3, 4]))

    def test_indices_on_edge(self):
        data = [np.array([0, 5, 20, 30, 40])]
        bins = [np.array([0, 10, 20, 30])]
        hist = HistWithIndex(data, bins)
        idx, _ = hist._get_indices_for_photons()
        np.testing.assert_equal(idx, np.array([1, 1, 3, 3, 4]))

    def test_multidim(self):
        data = [np.array([0, 5, 20, 30, 40]), np.array([100, 200, 200, 300, 400])]
        bins = [np.array([0, 10, 20, 30]), np.array([100, 150, 200])]
        hist = HistWithIndex(data, bins)
        idx, _ = hist._get_indices_for_photons()
        np.testing.assert_equal(idx, np.array([5, 6, 14, 15, 19]))

    def test_hist_populate_basic(self):
        data = [np.array([5, 15, 25])]
        bins = [np.array([0, 10, 20, 30])]
        hist = HistWithIndex(data, bins)
        hist.run()
        np.testing.assert_equal(hist.hist_photons, np.array([1, 1, 1]))

    def test_myhist_against_histdd(self):
        data = [np.array([5, 15, 25])]
        bins = [np.array([0, 10, 20, 30])]
        hist = HistWithIndex(data, bins)
        hist.run()
        np.testing.assert_equal(hist.hist_photons, np.histogramdd(data, bins)[0])

    def test_my_multidim_against_histdd(self):
        data = [
            np.array([0, 5, 20, 30, 40]),
            np.array([100, 200, 200, 300, 400]),
            np.array([13, 14, 15, 16, 16]),
        ]
        bins = [
            np.array([0, 10, 20, 30]),
            np.array([100, 150, 200]),
            np.array([10, 20]),
        ]
        hist = HistWithIndex(data, bins)
        hist.run()
        histdd = np.histogramdd(data, bins)[0]
        np.testing.assert_equal(hist.hist_photons, histdd)


class TestFlimCalc:
    """A test suite for the FLIM calculation."""

    def test_decay_borders_basic(self):
        hist = np.array([10, 20, 10, 5, 4, 3, 2, 10])
        peaks = np.array([1])
        props = {"peak_heights": np.array([20])}
        decay, _max, _min = find_decay_borders(hist, peaks, props)
        np.testing.assert_equal(decay, np.array([20, 10, 5, 4, 3, 2]))
        np.testing.assert_equal(_max, np.array([20]))
        np.testing.assert_equal(_min, np.array([2]))

    def test_decay_borders_min_at_end(self):
        hist = np.array([10, 20, 10, 5, 4, 3, 2])
        peaks = np.array([1])
        props = {"peak_heights": np.array([20])}
        decay, _max, _min = find_decay_borders(hist, peaks, props)
        np.testing.assert_equal(decay, np.array([20, 10, 5, 4, 3, 2]))
        np.testing.assert_equal(_max, np.array([20]))
        np.testing.assert_equal(_min, np.array([2]))

    def test_decay_borders_peak_at_start(self):
        hist = np.array([15, 10, 5, 4, 3, 2])
        peaks = np.array([0])
        props = {"peak_heights": np.array([15])}
        decay, _max, _min = find_decay_borders(hist, peaks, props)
        np.testing.assert_equal(decay, np.array([15, 10, 5, 4, 3, 2]))
        np.testing.assert_equal(_max, np.array([15]))
        np.testing.assert_equal(_min, np.array([2]))

    def test_decay_borders_wider_peak(self):
        hist = np.array([10, 19, 18, 15, 10, 5, 4, 3, 2, 10, 3])
        peaks = np.array([1])
        props = {"peak_heights": np.array([19])}
        decay, _max, _min = find_decay_borders(hist, peaks, props)
        np.testing.assert_equal(decay, np.array([19, 18, 15, 10, 5, 4, 3, 2]))
        np.testing.assert_equal(_max, np.array([19]))
        np.testing.assert_equal(_min, np.array([2]))

    def test_calc_lifetime(self):
        amp = 100
        tau = 35
        length = 125
        data = []
        for index in range(1, 126):
            data.extend([index for _ in range(1, int(amp * np.exp(-index / tau)) + 1)])
        tau = calc_lifetime(data)
        assert np.allclose([tau], [35], atol=0.5)

    def test_partition_pipe(self):
        amp = 100
        tau = 35
        length = 125
        num_of_bins = 100
        data_of_single_bin = []
        bin_idx = []
        for index in range(1, length + 1):
            data_of_single_bin.extend(
                [index for _ in range(1, int(amp * np.exp(-index / tau)) + 1)]
            )
        for pixel_idx in range(num_of_bins):
            bin_idx.extend([pixel_idx for _ in range(len(data_of_single_bin))])
        data_of_single_bin = num_of_bins * data_of_single_bin
        data_of_single_bin = np.array(data_of_single_bin)
        bin_idx = np.array(bin_idx)
        fl = FlimCalc(data_of_single_bin, bin_idx)
        fl._partition_photons_into_bins()
        true_taus = np.ones((num_of_bins)) * tau
        bins = np.arange(num_of_bins)
        assert np.allclose(fl.hist_arrivals["since_laser"], true_taus, atol=0.5)
        assert np.allclose(fl.hist_arrivals["bin"], bins)

    @pytest.mark.skip
    def test_normalization(self):
        arrivals = pd.DataFrame({"since_laser": np.array([0, 125 / 2, 125])})
        fl = FlimCalc(arrivals["since_laser"].to_numpy(), np.array([1, 2, 3]))
        fl.hist_arrivals = arrivals
        fl._normalize_taus()
        np.testing.assert_equal(
            fl.hist_arrivals["lifetime"], np.array([0, 127, 255], dtype=np.float32)
        )



import numpy as np
import pandas as pd
import attr
from attr.validators import instance_of
from enum import Enum, auto


class LineSignalType(Enum):
    """ All possible types of correct/incorrect line signals """

    VALID_UNIDIR = auto()
    VALID_BIDIR = auto()
    EXTRA_LINES = auto()
    SINGLE_PIXEL = auto()


@attr.s(slots=True)
class LineRectifier:
    """
    Receives the input lines and makes sure that they're OK and will render a good volume.

    Returns the np.array with the corrected lines.

    Errors - ``ValueError`` if signal can't be corrected.
    """

    lines = attr.ib(validator=instance_of(np.ndarray))  # raw line signal
    x_pixels = attr.ib(
        default=512, validator=instance_of(int)
    )  # number of expected lines
    bidir = attr.ib(default=False, validator=instance_of(bool))  # bidir or unidir scan
    ALLOWED_THRESHOLD = attr.ib(default=0.1, validator=instance_of(float))
    end_time = attr.ib(default=np.uint64(1), validator=instance_of(np.uint64))
    rectified_lines = attr.ib(init=False)  # returned signal
    __process_method = attr.ib(init=False)

    def rectify(self) -> np.ndarray:
        """ Main method to run the validation process on the line signal """
        signal_type = self.__assess_line_sig()
        self.rectified_lines = self.__process_method[signal_type]()
        return self.rectified_lines

    def __attrs_post_init__(self):
        """
        Generate a basic dictionary to choose the right processing method
        :return:
        """
        self.__process_method = {
            LineSignalType.EXTRA_LINES: self.__truncate_extra_lines,
            LineSignalType.VALID_UNIDIR: self.__preprocess_unidir_lines,
            LineSignalType.VALID_BIDIR: self.__preprocess_bidir_lines,
            LineSignalType.SINGLE_PIXEL: self.__single_pixel,
        }

    def __assess_line_sig(self) -> LineSignalType:
        """ Preliminary effort to detect the quality of the line signal """
        self.lines.sort()
        num_of_lines = len(self.lines)
        if num_of_lines > 1:
            if (
                np.abs(1 - (num_of_lines / self.x_pixels)) > self.ALLOWED_THRESHOLD
            ):  # line signal was too corrupt
                if (
                    num_of_lines >= self.x_pixels
                ):  # corrupt, but due to some extra lines
                    return LineSignalType.EXTRA_LINES
                else:  # really too corrupt
                    raise ValueError

            else:  # lines signal is corrupt, but we can save it
                return (
                    LineSignalType.VALID_BIDIR
                    if self.bidir
                    else LineSignalType.VALID_UNIDIR
                )

        else:  # single pixel frames, perhaps
            return LineSignalType.SINGLE_PIXEL

    def __truncate_extra_lines(self) -> np.ndarray:
        """ Called when the line signal is both dirty and contains extra lines """
        mean_diff = np.diff(self.lines).mean()
        return np.r_[
            self.lines[: self.x_pixels],
            np.array([self.lines[self.x_pixels - 1] + mean_diff], dtype=np.uint64),
        ]

    def __single_pixel(self) -> np.ndarray:
        """
        Dumbed down case of a single-pixel frame
        :return:
        """
        return np.linspace(
            self.lines[0], self.end_time, num=self.x_pixels + 1, endpoint=True
        )

    def __preprocess_unidir_lines(self) -> np.ndarray:
        """ Obtain the needed parameters for the rectify_lines method """
        lines = pd.Series(self.lines)
        diffs = lines.diff()
        rel_idx = np.where(
            np.abs(diffs.pct_change(periods=1)) > self.ALLOWED_THRESHOLD
        )[0]
        rel_diff = np.diff(rel_idx) == 2
        if rel_diff.shape[0] > 0 and sum(rel_diff) > 0:
            for idx in np.where(np.concatenate((rel_diff, np.array([False]))))[
                0
            ]:  # edge case
                rel_idx[idx + 1] -= 1
        mean_val = diffs.drop(rel_idx).mean()
        return self.__rectify_lines(
            lines=lines, diffs=diffs, mean_val=mean_val, rel_idx=rel_idx
        )

    def __preprocess_bidir_lines(self) -> np.ndarray:
        """ Obtain the needed parameters for the rectify_lines method """
        self.ALLOWED_THRESHOLD = 0.15  # x100 percent
        lines = pd.Series(self.lines)
        diff1 = lines.diff()
        diffs = lines.rolling(2).sum().diff()
        diffs[1] = diffs.mean()
        rel_idx = np.where(
            np.abs(diffs.pct_change(periods=1)) > self.ALLOWED_THRESHOLD
        )[0]
        if len(rel_idx) == 0:
            return self.__rectify_lines(
                lines=lines, diffs=diff1, mean_val=diff1.mean(), rel_idx=rel_idx
            )

        # if len(rel_idx) % 2 != 0:  # every misplaced line should have a start and an end
        #     raise ValueError
        idx_to_throw = np.array([], dtype=np.int64)
        if rel_idx.shape[0] > 1:
            for idx, _ in enumerate(rel_idx[::2]):
                try:
                    idx_to_throw = np.concatenate(
                        (
                            idx_to_throw,
                            np.arange(rel_idx[idx * 2], rel_idx[idx * 2 + 1] + 1),
                        )
                    )
                except IndexError:
                    idx_to_throw = np.concatenate((idx_to_throw, rel_idx[-1]))
        else:
            idx_to_throw = rel_idx.copy()

        final_rel_idx = rel_idx[::2]
        # Double jumps of rows missing
        rel_diff = np.diff(final_rel_idx) == 2
        if rel_diff.shape[0] > 0 and sum(rel_diff) > 0:
            for idx in np.where(np.concatenate((rel_diff, np.array([False]))))[
                0
            ]:  # edge case
                final_rel_idx[idx + 1] -= 1

        # Check for multiline double spacing
        double: np.ndarray = np.where(np.diff(rel_idx)[::2] == 4)[0]
        if len(double) > 0:
            for idx in double:
                indices_to_concat = [
                    place for place in range(rel_idx[idx] + 1, rel_idx[idx + 1])
                ]
                new_rel_idx = np.concatenate(
                    (
                        rel_idx[: idx + 1],
                        rel_idx[idx + 2 :],
                        np.array(indices_to_concat, dtype=np.int64),
                    )
                )
        else:
            new_rel_idx = sorted(rel_idx)[::2]
        final_rel_idx = np.unique(np.concatenate((final_rel_idx, new_rel_idx)))

        # Check for multiline triple spacing
        triple: np.ndarray = np.where(np.diff(rel_idx)[::2] == 5)[0]
        if len(triple) > 0:
            for idx in triple:
                indices_to_concat = [
                    place for place in range(rel_idx[idx] + 1, rel_idx[idx + 1])
                ]
                rel_idx = np.concatenate(
                    (rel_idx, np.array(indices_to_concat, dtype=np.int64))
                )
        else:
            rel_idx.sort()
        final_rel_idx = np.unique(np.concatenate((final_rel_idx, rel_idx)))

        mean_val = diff1.drop(idx_to_throw).mean()
        return self.__rectify_lines(
            lines=lines, diffs=diff1, mean_val=mean_val, rel_idx=final_rel_idx
        )

    def __rectify_lines(
        self, lines: pd.Series, diffs: pd.Series, mean_val: float, rel_idx: np.ndarray
    ) -> np.ndarray:
        """
        Rectify a semi-broken line signal.

        :return np.ndarray: Rectified lines
        """

        if rel_idx.shape[0] > 0.2 * self.x_pixels:  # too many missing lines
            raise ValueError

        if len(lines) - 1 in rel_idx:  # last lines are missing
            needed_lines = 1 + (self.x_pixels + 1 - (len(lines) + len(rel_idx)))
            if needed_lines < 1:
                raise ValueError
            lines = lines[:-1].append(
                pd.Series(
                    np.linspace(
                        start=lines.iloc[-2] + mean_val,
                        stop=lines.iloc[-2] + ((needed_lines + 1) * mean_val),
                        num=needed_lines,
                        dtype=np.uint64,
                    )
                )
            )
            rel_idx = rel_idx[rel_idx != len(diffs) - 1]
        if (
            np.abs((diffs.iloc[1] - mean_val) / mean_val) > self.ALLOWED_THRESHOLD
        ):  # first line came late
            rel_idx = np.r_[rel_idx, 1]
        if len(rel_idx) > 0:  # straight-forward addition of line signals
            for val in rel_idx:
                missing_lines = int(np.around(diffs[val] / mean_val)) - 1
                lines = lines.append(
                    pd.Series(
                        np.linspace(
                            start=lines.iloc[val - 1] + mean_val,
                            stop=lines.iloc[val],
                            endpoint=False,
                            num=missing_lines,
                            dtype=np.uint64,
                        )
                    )
                )

        if len(lines) != self.x_pixels:  # lines weren't recorded from the get-go
            if lines.iloc[0] == 0:  # we'll append at the end
                lines = lines.sort_values().reset_index(drop=True)
                needed_lines = np.abs(self.x_pixels - len(lines))
                lines = lines.append(
                    pd.Series(
                        np.linspace(
                            start=lines.iloc[-1] + mean_val,
                            stop=lines.iloc[-1] + (needed_lines + 1) * mean_val,
                            endpoint=False,
                            num=needed_lines,
                            dtype=np.uint64,
                        )
                    )
                )
            else:
                lines = lines.append(
                    pd.Series(
                        np.linspace(
                            start=0,
                            stop=lines.iloc[0],
                            endpoint=False,
                            num=np.abs(self.x_pixels - len(lines)),
                            dtype=np.uint64,
                        )
                    )
                )
        lines = lines.sort_values().reset_index(drop=True)
        lines = np.r_[
            lines.iloc[: self.x_pixels],
            np.array([lines.iloc[self.x_pixels - 1] + mean_val], dtype=np.uint64),
        ]
        return lines



import attr
from attr.validators import instance_of
import numpy as np
import pandas as pd


@attr.s(slots=True)
class Deinterleave:
    """
    **HIGHLY EXPERIMENTAL**

    This class can differentiate the data recorded in PMT1 into two
    distinct channels by observing the time that has passed between
    the laser pulse and the photon detection time. Early photons
    should be allocated to the first channel, while the latter ones
    will be moved into a new channel, "PMT7".

    This class should be used when two (or more, in the future) beams
    are co-aligned and excite the sample near-simultaneously.
    """

    photons = attr.ib(validator=instance_of(pd.DataFrame))
    reprate = attr.ib(default=80.3e6, validator=instance_of(float))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    num_of_beams = attr.ib(default=2, validator=instance_of(int))

    @property
    def bins_bet_pulses(self):
        return int(np.ceil(1 / self.reprate / self.binwidth))

    def run(self) -> pd.DataFrame:
        """
        Main pipeline for this class. Returns a DataFrame in which
        Channel 1 was deinterleaved into two channels, early and late
        with the late data going into channel 7.
        """
        late_photons_mask = self.photons.xs(1, level=0).mask(
            self.photons["time_rel_pulse"].xs(1, level=0) > (self.bins_bet_pulses // 2)
        )
        early_photons = late_photons_mask.dropna().assign(
            time_rel_pulse=lambda x: x.time_rel_pulse.astype(np.uint8)
        )
        late_photons = self.photons.xs(1, level=0).loc[
            late_photons_mask["time_rel_pulse"].isna(), :
        ]

        # TODO: Change following paragraph when pandas 0.24 hits with Int64 extension type
        for column in early_photons.columns:
            early_photons[column] = early_photons[column].astype(
                late_photons[column].dtype
            )
        remaining_data_channels = self.photons.index.levels[0].categories[1:-1]
        other_channels_photon_df = []
        for chan in remaining_data_channels:
            other_channels_photon_df.append(self.photons.xs(chan, level=0))

        new_photons = (
            pd.concat(
                [early_photons] + other_channels_photon_df + [late_photons],
                keys=self.photons.index.levels[0].categories.values,
                names=["Channel"],
            )
            .reset_index(level=0)
            .assign(Channel=lambda x: x["Channel"].astype("category"))
            .set_index("Channel", append=True)
            .reorder_levels((2, 0, 1))
        )
        return new_photons



from typing import Dict, Tuple, Union, Any
import logging
import re
from enum import Enum

from attr.validators import instance_of
import numpy as np
import attr

from pysight.ascii_list_file_parser import timepatch_switch
from pysight.ascii_list_file_parser.tabulation import Tabulate
from pysight.ascii_list_file_parser.distribute_data import DistributeData


NANOSECONDS_PER_FSTCHAN = 6.4e-9  # Multiscaler const parameter
MULTISCALER_RESOLUTION = 100e-12  # 100 picoseconds resolution
MULTISCALER_EOS_DEADTIME = 96  # 96 nanoseconds end-of-sweep deadtime


class LstFormat(Enum):
    ASCII = "ascii"
    BINARY = "binary"


@attr.s(slots=True)
class ReadMeta:
    """
    Manage pipeline of file IO process. Parses metadata, doesn't read the actual
    data in it.

    :param str filename: List file name
    :param str input_start: Data type in analog channel 'START' (6)
    :param str input_stop1: Data type in analog channel 'STOP1' (1)
    :param str input_stop2: Data type in analog channel 'STOP2' (2)
    :param float binwidth: Multiscaler resolution in seconds (100-800 picoseconds)
    :param bool use_sweeps: Use the sweeps counter as a new line indicator
    :param float mirror_phase: Phase delay for the line signal (in us)
    """

    filename = attr.ib(validator=instance_of(str))
    debug = attr.ib(default=False, validator=instance_of(bool))
    input_start = attr.ib(default="Frames", validator=instance_of(str))
    input_stop1 = attr.ib(default="PMT1", validator=instance_of(str))
    input_stop2 = attr.ib(default="Lines", validator=instance_of(str))
    input_stop3 = attr.ib(default="Empty", validator=instance_of(str))
    input_stop4 = attr.ib(default="Empty", validator=instance_of(str))
    input_stop5 = attr.ib(default="Empty", validator=instance_of(str))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    use_sweeps = attr.ib(default=False, validator=instance_of(bool))
    mirror_phase = attr.ib(default=-2.78, validator=instance_of(float))
    is_binary = attr.ib(init=False)
    timepatch = attr.ib(init=False)
    data_range = attr.ib(init=False)
    time_after = attr.ib(init=False)
    acq_delay = attr.ib(init=False)
    bitshift = attr.ib(init=False)
    start_of_data_pos = attr.ib(init=False)
    dict_of_input_channels = attr.ib(init=False)
    lst_metadata = attr.ib(init=False)
    fill_fraction = attr.ib(init=False)
    num_of_channels = attr.ib(init=False)
    help_dict = attr.ib(init=False)

    def run(self):
        """ Open file and find the needed parameters """
        self.determine_binary()
        metadata: str = self.__get_metadata()
        self.timepatch: str = self.get_timepatch(metadata)
        if self.timepatch == "3" and not self.is_binary:
            raise NotImplementedError(
                'Timepatch value "3" is currently not supported for hex files. Please message the package owner.'
            )
        self.data_range: int = self.get_range(metadata)

        self.start_of_data_pos: int = self.get_start_pos()
        self.time_after: int = self.__get_hold_after(cur_str=metadata)
        self.acq_delay: int = self.__get_fstchan(cur_str=metadata)
        self.bitshift: int = self.__get_bitshift(cur_str=metadata)
        self.fill_fraction: float = self.__calc_actual_fill_fraction()
        self.dict_of_input_channels, self.num_of_channels = self.find_active_channels(
            metadata
        )
        # Grab some additional metadata from the file, to be saved later
        self.__parse_extra_metadata(metadata)
        self.lst_metadata["data_range"] = str(self.data_range)

    def __attrs_post_init__(self):
        self.help_dict = {
            "1": (self.input_stop1, "001"),
            "2": (self.input_stop2, "010"),
            "3": (self.input_stop3, "011"),
            "4": (self.input_stop4, "100"),
            "5": (self.input_stop5, "101"),
            "6": (self.input_start, "110"),
        }

    def __get_metadata(self) -> str:
        """
        Read the file's metadata to be parsed by other functions.

        :return str: String with the .lst file's metadata
        """
        file_mode = "rb" if self.is_binary else "r"
        with open(self.filename, file_mode) as f:
            metadata = f.read(5000)

        return metadata

    @staticmethod
    def hex_to_bin_dict() -> Dict:
        """ Create a simple dictionary that maps a hex input into a 4 letter binary output. """
        diction = {
            "0": "0000",
            "1": "0001",
            "2": "0010",
            "3": "0011",
            "4": "0100",
            "5": "0101",
            "6": "0110",
            "7": "0111",
            "8": "1000",
            "9": "1001",
            "a": "1010",
            "b": "1011",
            "c": "1100",
            "d": "1101",
            "e": "1110",
            "f": "1111",
        }
        return diction

    def determine_binary(self):
        """
        Determine whether we're dealing with a binary or a hex file.
        """
        if self.filename == "":
            raise ValueError("No filename given.")

        try:
            with open(self.filename, "r") as f:
                txt = f.read(400)
        except FileNotFoundError:
            raise FileNotFoundError(f"File {self.filename} doesn't exist.")
        except UnicodeDecodeError:
            with open(self.filename, "rb") as f:
                txt = f.read(400).decode()
        except Exception:
            raise Exception(f"File {self.filename} read unsuccessfully.")

        reg = re.compile(r"\nmpafmt=(\w{3})")
        match = reg.findall(txt)
        if len(match) == 0:
            raise RuntimeError("Could not resolve file type.")
        if match[0] == "dat":
            self.is_binary = True
        elif match[0] == "asc":
            self.is_binary = False

    def get_range(self, cur_str) -> int:
        """
        Finds the "range" of the current file in the proper units
        :return int: range as defined by MCS6A
        """
        if self.filename == "":
            raise ValueError("No filename given.")

        if self.is_binary:
            format_str: Union[str, bytes] = rb"range=(\d+)"
        else:
            format_str = r"range=(\d+)"

        format_range = re.compile(format_str)
        range_lst = int(re.search(format_range, cur_str).group(1))

        return range_lst

    def get_timepatch(self, cur_str: str) -> str:
        """
        Get the time patch value out of of a list file.

        :param cur_str: Start of file to be analyzed.
        :return str: Time patch value as string.
        """
        if self.filename == "":
            raise ValueError("No filename given.")

        if self.is_binary:
            format_str: bytes = rb"time_patch=(\w+)"
        else:
            format_str: str = r"time_patch=(\w+)"

        format_timepatch = re.compile(format_str)
        timepatch = re.search(format_timepatch, cur_str).group(1)
        try:
            timepatch = timepatch.decode("utf-8")
        except AttributeError:
            assert isinstance(timepatch, str)
        finally:
            if self.is_binary and timepatch in ("1a", "2a", "22", "32", "2"):
                raise NotImplementedError(
                    f"The timepatch used ({timepatch}) isn't supported "
                    "for binary files since it uses a 6-byte word representation. "
                    "Please disallow this option in the MPANT software."
                )
            return timepatch

    def __get_bitshift(self, cur_str: str) -> int:
        """
        Finds and parses the bitshift in the current file's metadata.
        The bitshift is the power of two which represents the value,
        in time bins, of the "range" value of the list file.
        The bitshift is given in a hexadecimal format, so the function
        has to convert it to decimal aswell.
        :param cur_str: Text metadata of the current file.
        :return int: Bitshift value (2 ** bitshift)
        """
        if self.is_binary:
            format_str: bytes = rb"bitshift=(\w+)"
        else:
            format_str: str = r"bitshift=(\w+)"

        format_timepatch = re.compile(format_str)
        bitshift = re.search(format_timepatch, cur_str).group(1)
        try:
            bitshift = bitshift.decode("utf-8")
        except AttributeError:
            assert isinstance(bitshift, str)

        bitshift = 2 ** int(bitshift, 16)
        return bitshift

    def find_active_channels(self, cur_str) -> Tuple[Dict[str, str], int]:
        """
        Create a dictionary containing the active channels.

        :param cur_str: String to be analyzed.
        """
        if self.filename == "":
            raise ValueError("No filename given.")

        if self.is_binary:
            format_str = rb"\[CHN(\d)\].+?active=(\d)"
        else:
            format_str = r"\[CHN(\d)\].+?active=(\d)"

        format_active = re.compile(format_str, re.DOTALL)
        matches = format_active.findall(cur_str)
        dict_of_inputs = {}  # DataType: BinaryNumber (e.g. "Lines": "010")
        if self.is_binary:
            matches = [(match[0].decode(), match[1].decode()) for match in matches]
        for cur_match in matches[:-1]:
            if cur_match[1] == "1":  # channel is active, populate dict
                dict_of_inputs[self.help_dict[str(cur_match[0])][0]] = self.help_dict[
                    cur_match[0]
                ][1]
            elif (cur_match[1] == "0") and (
                self.help_dict[cur_match[0]][0] != "Empty"
            ):  # Inactive channel accroding to the multiscaler, but was marked as active by the user
                raise UserWarning(
                    f"Channel {cur_match[0]} didn't record data but was marked as active by the user."
                )

        if matches[-1][1] == "1":
            dict_of_inputs[self.help_dict["6"][0]] = self.help_dict["6"][1]
        elif (matches[-1][0] == "0") and (self.help_dict["6"][0] != "Empty"):
            raise UserWarning(
                "Channel 6 (START) didn't record data but was marked as active by the user."
            )

        num_of_channels = sum([1 for key in dict_of_inputs if "PMT" in key])

        assert len(dict_of_inputs) >= 1
        if "Empty" in dict_of_inputs.keys():
            logging.warning(
                f"At least one channel ({dict_of_inputs['Empty']})"
                " contained recorded data but was marked as 'Empty'."
            )
            dict_of_inputs.pop("Empty")

        return dict_of_inputs, num_of_channels

    def get_start_pos(self) -> int:
        """
        Returns the start position of the data

        :return int: Integer of file position for f.seek() method
        """
        if self.filename == "":
            raise ValueError("No filename given.")

        if self.is_binary:
            format_str = rb"DATA]\r\n"
            file_mode = "rb"
        else:
            format_str = r"DATA]\n"
            file_mode = "r"

        format_data = re.compile(format_str)
        pos_in_file: int = 0
        line_num: int = 0
        with open(self.filename, file_mode) as f:
            while pos_in_file == 0 and line_num < 1000:
                line = f.readline()
                line_num += 1
                match = re.search(format_data, line)
                if match is not None:
                    pos_in_file = f.tell()
                    return pos_in_file  # to have the [DATA] as header
        return -1

    def __get_hold_after(self, cur_str) -> int:
        """
        Read the time (in ns, and convert to timebins) that is considered a
        "hold after", or "hold-off", after a single sweep. Add that time, along with a 96 ns
        inherit delay, to all future times of the sweep.

        :param cur_str: String to parse
        :return int: Final time that has to be added to all sweeps, in timebins
        """
        if self.is_binary:
            format_str: bytes = rb"holdafter=([\w\+]+)"
        else:
            format_str: str = r"holdafter=([\w\+]+)"

        format_holdafter = re.compile(format_str)
        holdafter = int(re.search(format_holdafter, cur_str).group(1), 16)

        time_after_sweep = int(
            ((MULTISCALER_EOS_DEADTIME + (holdafter * 6.4)) * 10 ** (-9))
            / MULTISCALER_RESOLUTION
        )
        return time_after_sweep

    def __get_fstchan(self, cur_str: str) -> int:
        """
        Read the acquisition delay of each sweep, called "fstchan" in the list files

        :param str cur_str: Metadata to be parsed
        :return int: Acq delay in timebins
        """
        if self.is_binary:
            format_str: Union[bytes, str] = rb"fstchan=(\w+)"
        else:
            format_str = r"fstchan=(\w+)"

        format_fstchan = re.compile(format_str)
        fstchan = (
            int(re.search(format_fstchan, cur_str).group(1), 16)
            * NANOSECONDS_PER_FSTCHAN
        )  # in nanoseconds
        acq_delay = int(fstchan / MULTISCALER_RESOLUTION)
        return acq_delay

    def __parse_extra_metadata(self, metadata):
        """
        Update self.lst_metadata with some additional information

        :param metadata: Data from the start of the lst file.
        """
        list_to_parse = [
            "fstchan",
            "holdafter",
            "periods",
            "rtpreset",
            "cycles",
            "sequences",
            "range",
            "sweepmode",
            "fdac",
            "swpreset",
        ]
        self.lst_metadata = {}
        for cur_str in list_to_parse:
            self.__parse_str(metadata, cur_str)

        self.lst_metadata["mirror_phase"] = str(self.mirror_phase)

    def __parse_str(self, metadata, str_to_parse):
        """
        Find str_to_parse in metadata and place the corresponding value in
        the dictionary self.lst_metadata
        """
        if self.is_binary:
            self.__parse_str_bin(metadata, str_to_parse)
        else:
            self.__parse_str_ascii(metadata, str_to_parse)

    def __parse_str_bin(self, metadata, str_to_parse):
        """Parse a binary file for metadata."""
        format_str: bytes = str_to_parse.encode("utf-8") + rb"=(\w+)"
        format_regex = re.compile(format_str)
        try:
            self.lst_metadata[str_to_parse] = (
                re.search(format_regex, metadata).group(1).decode()
            )
        except AttributeError:  # field is non-existent
            pass

    def __parse_str_ascii(self, metadata, str_to_parse):
        """Parse an ASCII file for metadata."""
        format_str: str = str_to_parse + r"=(\w+)"
        format_regex = re.compile(format_str)
        try:
            self.lst_metadata[str_to_parse] = re.search(format_regex, metadata).group(1)
        except AttributeError:  # field is non-existent
            pass

    def __calc_actual_fill_fraction(self) -> float:
        """
        If we're using sweeps as lines then the true fill fraction is determined
        by the multiscaler's parameters, like hold_after and acquisiton delay.
        :return float: True fill fraction
        """
        if self.use_sweeps:
            fill_frac = self.data_range / (
                self.acq_delay + self.data_range + self.time_after
            )
            return fill_frac * 100  # in percent
        else:
            return -1.0


def ascii_parsing(cur_file: ReadMeta, raw_data: np.ndarray, config: Dict[str, Any]):
    """Preliminary readout of data from ASCII file."""
    dict_of_slices_hex = timepatch_switch.ChoiceManagerHex().process(cur_file.timepatch)
    tabulated_data = Tabulate(
        data_range=cur_file.data_range,
        data=raw_data,
        dict_of_inputs=cur_file.dict_of_input_channels,
        use_tag_bits=config["tagbits"]["tag_bits"],
        dict_of_slices_hex=dict_of_slices_hex,
        bitshift=cur_file.bitshift,
        time_after_sweep=cur_file.time_after,
        acq_delay=cur_file.acq_delay,
        num_of_channels=cur_file.num_of_channels,
    )
    tabulated_data.run()

    del raw_data
    separated_data = DistributeData(
        df=tabulated_data.df_after_timepatch,
        dict_of_inputs=tabulated_data.dict_of_inputs,
        use_tag_bits=config["tagbits"]["tag_bits"],
    )
    separated_data.run()
    return separated_data.data_to_grab, separated_data.dict_of_data



"""
__author__ = Hagai Hargil
"""

from unittest import TestCase
from pysight.nd_hist_generator.line_signal_validators.validation_tools import *
import pandas as pd
import numpy as np


class TestFrame(TestCase):
    """
    Tests for the validation functions
    """

    dict_of_data = dict(
        PMT1=pd.DataFrame([1, 10, 20, 30], columns=["abs_time"]),
        Lines=pd.DataFrame([0, 5, 10, 15, 20, 25, 30, 35], columns=["abs_time"]),
    )
    vlad = SignalValidator(dict_of_data)

    def test_last_event_only_pmt(self):
        dict_of_data = {"PMT1": pd.DataFrame([1, 5, 3], columns=["abs_time"])}
        last = SignalValidator(dict_of_data)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 5)

    def test_last_event_two_pmts(self):
        dict_of_data = {
            "PMT1": pd.DataFrame([1, 5, 3], columns=["abs_time"]),
            "PMT2": pd.DataFrame([1, 2, 3], columns=["abs_time"]),
        }
        last = SignalValidator(dict_of_data)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 5)

    def test_last_event_with_frames(self):
        frame_data = pd.DataFrame([0, 100], columns=["abs_time"])
        dict_of_data = {
            "PMT1": pd.DataFrame([1, 2, 3], columns=["abs_time"]),
            "Frames": frame_data,
            "Lines": [1, 2, 3],
        }
        last = SignalValidator(dict_of_data)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 200)

    def test_last_event_with_single_frame(self):
        frame_data = pd.DataFrame([100], columns=["abs_time"])
        dict_of_data = {
            "PMT1": pd.DataFrame([1, 2, 3], columns=["abs_time"]),
            "Frames": frame_data,
            "Lines": [1, 2, 3],
        }
        last = SignalValidator(dict_of_data)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 200)

    def test_last_event_with_lines_less_than_needed_single_frame(self):
        line_data = pd.DataFrame([0, 10, 20], columns=["abs_time"])
        dict_of_data = {
            "PMT1": pd.DataFrame([1, 2, 3], columns=["abs_time"]),
            "Lines": line_data,
        }
        last = SignalValidator(
            dict_of_data, num_of_lines=5
        )._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 50)

    def test_last_event_with_lines_less_than_needed_more_frames(self):
        line_data = pd.DataFrame([0, 10, 20, 30, 40], columns=["abs_time"])
        dict_of_data = {
            "PMT1": pd.DataFrame([1, 2, 3], columns=["abs_time"]),
            "Lines": line_data,
        }
        last = SignalValidator(
            dict_of_data, num_of_lines=3
        )._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 30)

    def test_last_event_with_lines_just_like_needed(self):
        line_data = pd.DataFrame([0, 10, 20], columns=["abs_time"])
        dict_of_data = {
            "PMT1": pd.DataFrame([1, 2, 3], columns=["abs_time"]),
            "Lines": line_data,
        }
        last = SignalValidator(
            dict_of_data, num_of_lines=3
        )._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 30)

    def test_last_event_with_lines_more_than_needed(self):
        line_data = pd.DataFrame(np.arange(0, 100, 10), columns=["abs_time"])
        dict_of_data = {
            "PMT1": pd.DataFrame([1, 2, 3], columns=["abs_time"]),
            "Lines": line_data,
        }
        last = SignalValidator(
            dict_of_data, num_of_lines=3
        )._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 90)

    def test_bins_bet_lines(self):
        line_freq = 10.0
        binwidth = 0.01
        bidir = False
        bins = SignalValidator(
            dict_of_data=self.dict_of_data,
            line_freq=line_freq,
            binwidth=binwidth,
            bidir=bidir,
        )._SignalValidator__bins_bet_lines()
        self.assertEqual(10, bins)

    def test_bins_bet_lines_bidir(self):
        line_freq = 10.0
        binwidth = 0.01
        bidir = True
        bins = SignalValidator(
            dict_of_data=self.dict_of_data,
            line_freq=line_freq,
            binwidth=binwidth,
            bidir=bidir,
        )._SignalValidator__bins_bet_lines()
        self.assertEqual(5, bins)

    def test_extrapolate_without_zero(self):
        sigval = SignalValidator(
            self.dict_of_data, line_freq=0.5, binwidth=1.0, delay_between_frames=0.0
        )
        line_point = self.dict_of_data["Lines"].iat[1, 0]
        sigval.last_event_time = 15
        sigval.line_delta = sigval._SignalValidator__bins_bet_lines()
        sigval.num_of_lines = int(sigval.last_event_time // sigval.line_delta)
        returned_lines = sigval._SignalValidator__extrapolate_line_data(
            line_point=line_point
        )
        real_lines = np.arange(1, 15, step=2, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines.abs_time.tolist(), real_lines.tolist())

    def test_extrapolate_from_zero(self):
        sigval = SignalValidator(
            self.dict_of_data, line_freq=1.0, binwidth=1.0, delay_between_frames=0.0
        )
        line_point = self.dict_of_data["Lines"].iat[2, 0]
        sigval.last_event_time = 15
        sigval.line_delta = sigval._SignalValidator__bins_bet_lines()
        sigval.num_of_lines = int(sigval.last_event_time // sigval.line_delta)
        returned_lines = sigval._SignalValidator__extrapolate_line_data(
            line_point=line_point
        )
        real_lines = np.arange(0, 15, step=1, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines.abs_time.tolist(), real_lines.tolist())

    def test_extrapolate_without_timepoint(self):
        sigval = SignalValidator(
            self.dict_of_data, line_freq=1.0, binwidth=1.0, delay_between_frames=0.0
        )
        sigval.last_event_time = 15
        sigval.line_delta = sigval._SignalValidator__bins_bet_lines()
        sigval.num_of_lines = int(sigval.last_event_time // sigval.line_delta)
        returned_lines = sigval._SignalValidator__extrapolate_line_data()
        real_lines = np.arange(0, 15, step=1, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines.abs_time.tolist(), real_lines.tolist())

    def test_extrapolate_without_timepoint_and_zero(self):
        sigval = SignalValidator(
            self.dict_of_data, line_freq=0.5, binwidth=1.0, delay_between_frames=0.0
        )
        sigval.last_event_time = 15
        sigval.line_delta = sigval._SignalValidator__bins_bet_lines()
        sigval.num_of_lines = int(sigval.last_event_time // sigval.line_delta) + 1
        returned_lines = sigval._SignalValidator__extrapolate_line_data()
        real_lines = np.arange(0, 15, step=2, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines.abs_time.tolist(), real_lines.tolist())

    def test_pipeline_single_frame_si(self):
        lines = [0, 5, 10, 15, 20, 25, 30, 70, 75]
        dict_of_data = dict(
            PMT1=pd.DataFrame(
                [1, 10, 20, 30, 100], columns=["abs_time"], dtype=np.uint64
            ),
            Lines=pd.DataFrame(lines, columns=["abs_time"], dtype=np.uint64),
        )
        binwidth = 1.0
        line_freq = 0.2
        delay_bet_frames = 0.0
        num_of_lines = 7
        sigval = SignalValidator(
            dict_of_data=dict_of_data,
            binwidth=binwidth,
            line_freq=line_freq,
            delay_between_frames=delay_bet_frames,
            num_of_lines=num_of_lines,
            image_soft=ImagingSoftware.SCANIMAGE.value,
        )
        sigval.run()
        self.assertSequenceEqual(
            sigval.dict_of_data["Lines"].abs_time.tolist(), lines[:-2]
        )
        self.assertSequenceEqual(sigval.dict_of_data["Frames"].abs_time.tolist(), [0])
        self.assertEqual(sigval.line_delta, 5)

    def test_pipeline_few_frames_si(self):
        lines = [0, 5, 10, 45, 50, 55, 90, 95, 100, 135, 140]
        dict_of_data = dict(
            PMT1=pd.DataFrame(
                [1, 10, 20, 30, 100], columns=["abs_time"], dtype=np.uint64
            ),
            Lines=pd.DataFrame(lines, columns=["abs_time"], dtype=np.uint64),
        )
        binwidth = 1.0
        line_freq = 0.2
        delay_bet_frames = 0.0
        num_of_lines = 3
        sigval = SignalValidator(
            dict_of_data=dict_of_data,
            binwidth=binwidth,
            line_freq=line_freq,
            delay_between_frames=delay_bet_frames,
            num_of_lines=num_of_lines,
            image_soft=ImagingSoftware.SCANIMAGE.value,
        )
        sigval.run()
        self.assertSequenceEqual(
            sigval.dict_of_data["Lines"].abs_time.tolist(), lines[:-2]
        )
        self.assertSequenceEqual(
            sigval.dict_of_data["Frames"].abs_time.tolist(), [0, 45, 90]
        )
        self.assertEqual(sigval.line_delta, 5)

    def test_pipeline_single_frame_mscan(self):
        lines = [0, 5, 10, 15, 20, 25, 30, 35, 40]
        dict_of_data = dict(
            PMT1=pd.DataFrame(
                [1, 10, 20, 30, 100], columns=["abs_time"], dtype=np.uint64
            ),
            Lines=pd.DataFrame(lines, columns=["abs_time"], dtype=np.uint64),
        )
        binwidth = 1.0
        line_freq = 0.2
        delay_bet_frames = 0.0
        num_of_lines = 7
        sigval = SignalValidator(
            dict_of_data=dict_of_data,
            binwidth=binwidth,
            line_freq=line_freq,
            delay_between_frames=delay_bet_frames,
            num_of_lines=num_of_lines,
            image_soft=ImagingSoftware.MSCAN.value,
        )
        sigval.run()
        self.assertSequenceEqual(
            sigval.dict_of_data["Lines"].abs_time.tolist(), lines[:-2]
        )
        self.assertSequenceEqual(sigval.dict_of_data["Frames"].abs_time.tolist(), [0])
        self.assertEqual(sigval.line_delta, 5)

    def test_pipeline_few_frames_mscan(self):
        lines = [0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60]
        dict_of_data = dict(
            PMT1=pd.DataFrame(
                [1, 10, 20, 30, 100, 200], columns=["abs_time"], dtype=np.uint64
            ),
            Lines=pd.DataFrame(lines, columns=["abs_time"], dtype=np.uint64),
        )
        binwidth = 1.0
        line_freq = 0.2
        delay_bet_frames = 0.0
        num_of_lines = 3
        sigval = SignalValidator(
            dict_of_data=dict_of_data,
            binwidth=binwidth,
            line_freq=line_freq,
            delay_between_frames=delay_bet_frames,
            num_of_lines=num_of_lines,
            image_soft=ImagingSoftware.MSCAN.value,
        )
        sigval.run()
        self.assertSequenceEqual(
            sigval.dict_of_data["Lines"].abs_time.tolist(), lines[:-1]
        )
        self.assertSequenceEqual(
            sigval.dict_of_data["Frames"].abs_time.tolist(), [0, 15, 30, 45]
        )
        self.assertEqual(sigval.line_delta, 5)



"""
Methods to calculate the lifetime of the generated
stacks
"""
from typing import Union, Optional, Tuple, Iterator, Any
import logging
import itertools

import numpy as np
import scipy.optimize
import attr
from attr.validators import instance_of, in_

from pysight.nd_hist_generator.outputs import PySightOutput


def in_channel(instance, attribute, value):
    if value not in instance.data.available_channels:
        return ValueError


@attr.s
class LifetimeCalc:
    """
    Using a generated :class:`pysight.nd_hist_generator.outputs.PySightOutput` object, calculates the
    total lifetime in that stack.

    :param `pysight.nd_hist_generator.outputs.PySightOutput` data: All data as generated by PySight.
    :param int channel: Channel on which to perform lifetime\
        calculation.
    :param int num_of_bins: Number of bins expected in the \
        lifetime decay curve.
    :param int bin_factor: By how much to bin the pixels\
        of the image. Increase when photon counts are low.
    """

    data = attr.ib(validator=instance_of(PySightOutput))
    channel = attr.ib(validator=[instance_of(int), in_channel])
    num_of_bins = attr.ib(default=125, validator=instance_of(int))
    bin_factor = attr.ib(default=2, validator=instance_of(int))
    raw_arrival_times = attr.ib(init=False)
    hist = attr.ib(init=False)
    params = attr.ib(init=False)
    cov = attr.ib(init=False)

    def fit_entire_fov(self) -> Optional["ExpFitParams"]:
        """
        Main method for class. Returns an ExpFitParams object
        which contains the result of the fitting done.
        """
        self.raw_arrival_times = self._get_photons()
        self.hist = self._gen_aligned_hist(self.raw_arrival_times)
        try:
            lowest_bin = self._get_low_bin_idx(self.hist)
        except ValueError:  # no valid lowest bin
            logging.warning("No valid decay curve found in the histogram.")
            return
        chopped_hist = self.hist[:lowest_bin]
        self.params, self.cov = self._fit_decay(chopped_hist)
        return ExpFitParams(self.params[0], self.params[1], self.params[2])

    def flim_binned(self) -> np.ndarray:
        """
        Create an image of the FOV colored by the lifetime of each
        2x2 square in the image.
        """
        channel_data = getattr(self.data, f"ch{self.channel}")
        shape_x = channel_data.shape[0]
        shape_y = channel_data.shape[1]
        assert shape_x == shape_y  # supports only square images ATM
        slices = self._gen_slices(shape_x)
        flim_img = self._populate_binned_image(slices, channel_data)
        return flim_img

    def _get_photons(self):
        """
        Returns an array of the photon arrival times relative to the
        last laser pulse.
        """
        return self.data.photons.xs(
            self.channel, level="Channel"
        ).time_rel_pulse.to_numpy()

    def _create_indices_per_ax(self, start, length=512) -> Tuple[Iterator[Any], ...]:
        """
        Create a vector which contains the indices at which we'll cut
        the original data in order to bin it.
        """
        idx = np.arange(start, length)
        idx = np.concatenate([None], idx, [None])
        idx2 = itertools.tee(idx)
        next(idx2[1])
        return idx2

    def _gen_slices(self, shape):
        """
        Create slice objects that will slice up the original
        array into a binned image.
        Each element of the returned list is a length 3 tuple, each
        of its elements being a slice, and each such slice details
        the starting and ending index of that bin.
        """

        idx_x = self._create_indices_per_ax(self.bin_factor, length=shape)
        idx_y = self._create_indices_per_ax(self.bin_factor, length=shape)
        xs = [slice(start, end) for start, end in zip(idx_x[0], idx_x[1])]
        ys = [slice(start, end) for start, end in zip(idx_y[0], idx_y[1])]
        return [(x, y, slice(None)) for x, y in zip(xs, ys)]

    def _populate_binned_image(self, slices, img):
        """
        Creates a binned image which contains the FLIM data from each of the
        original pixels of that bin.
        """
        flim_img = img.copy()
        for sl in slices:
            cur_data = img[sl].sum(axis=(0, 1))
            flim_img[sl] = self._fit_decay(cur_data)[0][1]  # computed tau
        return flim_img

    def _gen_aligned_hist(self, photons):
        """ Creates a histogram from the photon list and then moves it so that the highest
        bin is in the start of it.
        """
        hist, bins = np.histogram(photons, bins=self.num_of_bins)
        hist = np.roll(hist, -hist.argmax())
        return hist

    def _get_low_bin_idx(self, hist):
        """
        Finds in the histogram the edge of the decay curve, i.e.
        the lowest bin.
        """
        lowest = hist.argmin()
        if lowest < 2:
            raise ValueError
        return lowest

    def _fit_decay(self, y):
        """
        Use a curve fitting method to find the parameters of the
        exponential decay of the data
        """
        xdata = np.arange(len(y)) / (self.data.config["binwidth"] * 10 ** 9)
        popt, pcov = scipy.optimize.curve_fit(
            self.single_exp, xdata, y, (y.max(), 4, np.median(y))
        )
        return popt, pcov

    @staticmethod
    def single_exp(x, amp, tau, c):
        """
        A single decaying exponential fit, to be used with
        SciPy's ``curve_fit`` function.
        :param np.ndarray x: "Time" vector
        :param Union[float,int] amp: Starting amplitude
        :param Union[float,int] tau: Expected lifetime
        :param Union[float,int] c: Baseline level of the function
        """
        return amp * np.exp(-x / tau) + c


@attr.s(frozen=True)
class ExpFitParams:
    """
    Returned parameters from fitting a decay curve to a
    single exponential decay function.

    Parameters:
        :param float amplitude: Initial amplitude of decay.
        :param float tau: Fluorophore lifetime.
        :param float c: Constant offset of the decay above baseline.
    """

    amplitude = attr.ib(validator=instance_of(float))
    tau = attr.ib(validator=instance_of(float))
    c = attr.ib(validator=instance_of(float))



from unittest import TestCase
from pysight.nd_hist_generator.line_signal_validators.validation_tools import (
    SignalValidator,
)
from pysight.nd_hist_generator.line_signal_validators.mscan import MScanLineValidator
import pandas as pd
import numpy as np


class TestMScan(TestCase):
    """
    Tests for the ScanImageValidator class
    """

    dict_of_data = dict(
        PMT1=pd.DataFrame([1, 10, 20, 30], columns=["abs_time"]),
        Lines=pd.DataFrame([0, 5, 10, 15, 20, 25, 30, 35]),
        columns=["abs_time"],
    )
    vlad = SignalValidator(dict_of_data)



import logging
from typing import Dict, Tuple, List

import pandas as pd
import numpy as np
import attr
from attr.validators import instance_of
from scipy.optimize import curve_fit
from scipy.signal import find_peaks

from pysight.nd_hist_generator.tag_lens import TagPipeline


@attr.s(slots=True)
class Allocate(object):
    """
    Allocate photons to their coordinates (frames, lines, etc.).

    :param pd.DataFrame df_photons: DataFrame with photons
    :param dict dict_of_data: Organized data parsed from file
    :param float laser_freq: Frequency of laser pulses in Hz
    :param float binwidth: Binwidth of multiscaler in seconds (100 ps == 100e-12)
    :param bool bidir: Whether the scan was performed in a bi-directional manner
    :param float tag_freq: TAG lens frequency in Hz
    :param int tag_pulses: Number of TAG driver pulses per period (currently ``NotImplemented``)
    :param float phase: Phase of the resonant scanner when used in a bi-directional mode
    :param bool keep_unidir: Whether to keep photons from the returning phase of the resonant scanner
    :param bool flim: Whether to perform FLIM analysis (currently ``NotImplemented``)
    :param dict exp_params: Parameters from the FLIM fit (currently ``NotImplemented``)
    :param bool censor: Whether to perform censor correction (currently ``NotImplemented``)
    :param bool tag_interp_ok: Whether the TAG lens interpolation process completed successfully
    :param bool tag_to_phase: Whether to interpolate the TAG lens sinusoidal pattern
    :param int tag_offset: Offset in degrees of the TAG pulse
    """

    # TODO: Variable documentation
    df_photons = attr.ib(validator=instance_of(pd.DataFrame))
    dict_of_data = attr.ib(validator=instance_of(dict))
    laser_freq = attr.ib(default=80.3e6, validator=instance_of(float))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    bidir = attr.ib(default=False, validator=instance_of(bool))
    tag_freq = attr.ib(default=189e3, validator=instance_of(float))
    tag_pulses = attr.ib(default=1, validator=instance_of(int))
    phase = attr.ib(default=-2.78, validator=instance_of(float))
    keep_unidir = attr.ib(default=False, validator=instance_of(bool))
    flim = attr.ib(default=False, validator=instance_of(bool))
    exp_params = attr.ib(factory=dict, validator=instance_of(dict))
    censor = attr.ib(default=False, validator=instance_of(bool))
    tag_interp_ok = attr.ib(default=False, validator=instance_of(bool))
    tag_to_phase = attr.ib(default=True, validator=instance_of(bool))
    tag_offset = attr.ib(default=0, validator=instance_of(int))
    deinterleave = attr.ib(default=False, validator=instance_of(bool))
    sorted_indices = attr.ib(init=False)
    bins_for_flim = attr.ib(init=False)

    def run(self):
        """ Pipeline of allocation """
        logging.info(
            "Channels of events found. Allocating photons to their frames and lines..."
        )
        self.__allocate_photons()
        self.__allocate_tag()
        if self.flim or self.deinterleave:
            self.df_photons, rel_time = self.__interpolate_laser(self.df_photons)
            if self.flim or self.censor:
                self._rotate_laser_timings()
        # Censor correction addition:
        if "Laser" not in self.dict_of_data.keys():
            self.dict_of_data["Laser"] = 0
        self.__reindex_dict_of_data()

        logging.info("Relative times calculated. Creating Movie object...")

    @property
    def num_of_channels(self):
        return sum([1 for key in self.dict_of_data.keys() if "PMT" in key])

    @property
    def hist_bins_between_laser_pulses(self):
        return int(np.ceil((1 / self.laser_freq) * 1e9))

    def __allocate_photons(self):
        """
        Returns a dataframe in which each photon has an address as its index - its corresponding
        frame, line, and possibly laser pulse. This address is used to histogram it in the right pixel.
        The arrival time of each photon is calculated relative to this address - start-of-line, for example.
        The TAG lens address is different - each photon simply receives a phase between 0 and 2π.

        The function doesn't return, but it populates self.df_photons with the allocated data.
        """
        irrelevant_keys = {"PMT1", "PMT2", "TAG Lens"}
        relevant_keys = set(self.dict_of_data.keys()) - irrelevant_keys
        # Preparations
        column_heads = {"Lines": "time_rel_line_pre_drop", "Laser": "time_rel_pulse"}

        # Main loop - Sort lines and frames for all photons and calculate relative time
        for key in reversed(sorted(relevant_keys)):
            self.dict_of_data[key].sort_values("abs_time", inplace=True)
            sorted_indices = (
                np.digitize(
                    self.df_photons.loc[:, "abs_time"].values,
                    self.dict_of_data[key].loc[:, "abs_time"].values,
                )
                - 1
            )
            self.df_photons[key] = (
                self.dict_of_data[key].iloc[sorted_indices, 0].values
            )  # columns 0 is abs_time,
            # but the .iloc method is amazingly faster than .loc
            positive_mask = sorted_indices >= 0
            # drop photons that came before the first line
            self.df_photons = self.df_photons.iloc[positive_mask].copy()
            # relative time of each photon in accordance to the line\frame\laser pulse
            if "Frames" != key:
                self.df_photons[column_heads[key]] = (
                    self.df_photons["abs_time"] - self.df_photons[key]
                )
            self.sorted_indices = sorted_indices[sorted_indices >= 0]
            if "Lines" == key:
                self.__rectify_photons_in_uneven_lines()

            self.df_photons.set_index(keys=key, inplace=True, append=True, drop=True)

        assert len(self.df_photons) > 0
        assert np.all(self.df_photons.iloc[:, 0].values >= 0)  # finds NaNs as well
        self.df_photons.sort_index(
            level=self.df_photons.index.names, axis=0, inplace=True, sort_remaining=True
        )
        self.df_photons.sort_index(
            level=self.df_photons.index.names, axis=1, inplace=True, sort_remaining=True
        )
        assert self.df_photons.index.is_lexsorted()

    def __allocate_tag(self):
        """ Allocate photons to TAG lens phase """
        try:
            tag = self.dict_of_data["TAG Lens"].loc[:, "abs_time"]
        except KeyError:
            return
        else:
            logging.info("Interpolating TAG lens data...")
            tag_pipe = TagPipeline(
                photons=self.df_photons,
                tag_pulses=tag,
                freq=self.tag_freq,
                binwidth=self.binwidth,
                num_of_pulses=self.tag_pulses,
                to_phase=self.tag_to_phase,
                offset=self.tag_offset,
            )
            tag_pipe.run()
            self.df_photons = tag_pipe.photons
            self.tag_interp_ok = tag_pipe.finished_pipe

            logging.info("TAG lens interpolation finished.")

    def __requires_censoring(self, data: np.ndarray) -> bool:
        """
        Method to determine if we should undergo the censor correction process

        :param np.ndarray data: Bins of histogram from their peak onward.
        :return bool:
        """
        diffs = np.diff(data)
        if len(diffs) == 0:
            return True
        if np.all(diffs <= 0):  # No censoring occurred
            return False
        else:
            first_idx = np.argwhere(diffs >= 0)[0][0]
            diffs2 = diffs[first_idx:]
            if np.all(diffs2 >= 0) or np.all(diffs2 <= 0):
                return True
            else:  # either false alarm, or a third photon is on its way
                self.__requires_censoring(diffs2)
                return True

    def __interpolate_laser(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, List]:
        """
        Assign a time relative to a laser pulse for each photon. Assuming that the clock is synced to a 10 MHz signal.

        :param pd.DataFrame df: Dataframe with data for each photon.
        :return: Modified dataframe and the relative times.
        """
        needed_bins = self._find_integer_gcd()

        rel_time = []
        for chan in range(1, self.num_of_channels + 1):
            rel_time_first = (
                df.xs(key=chan, level="Channel", drop_level=False)["abs_time"].values
                % needed_bins
            )
            rel_time_per_pulse = rel_time_first % np.ceil(
                1 / (self.binwidth * self.laser_freq)
            ).astype(np.uint8)
            rel_time.append(rel_time_per_pulse)
            df.loc[chan, "time_rel_pulse"] = rel_time_per_pulse

        df["time_rel_pulse"] = np.uint8(df["time_rel_pulse"])
        return df, rel_time

    def _find_integer_gcd(self):
        """
        Converts the binwidth and laser frequency to the interger
        number of bins between pulses.
        """
        laser_freq = np.around(self.laser_freq, decimals=-6)  # MHz reprates
        bins_bet_pulses = np.around(laser_freq / self.binwidth, decimals=3)
        ns_between_pulses = (1 / laser_freq) * 1e9

        while int(bins_bet_pulses) != bins_bet_pulses:
            bins_bet_pulses *= 10

        while int(ns_between_pulses) != ns_between_pulses:
            ns_between_pulses *= 10

        return np.gcd(int(bins_bet_pulses), int(ns_between_pulses))

    def _find_num_pulses(self, needed_bins) -> int:
        period = int(needed_bins * self.binwidth * 1e9)
        laser_freq = np.around(self.laser_freq, decimals=-6)  # MHz reprates
        ns_between_pulses = (1 / laser_freq) * 1e9
        return int(period / ns_between_pulses)

    def __add_phase_offset_to_bidir_lines(self):
        """
        Uneven lines in a bidirectional scanning mode have to be offsetted.
        """
        phase_in_seconds = self.phase * 1e-6
        self.dict_of_data["Lines"].abs_time.iloc[1::2] -= np.uint64(
            phase_in_seconds / self.binwidth
        )

    def __rectify_photons_in_uneven_lines(self):
        """
        "Deal" with photons in uneven lines. Unidir - if keep_unidir is false, will throw them away.
        Bidir = flips them over (in the Volume object)
        """
        uneven_lines = np.remainder(self.sorted_indices, 2)
        if self.bidir:
            self.df_photons.rename(
                columns={"time_rel_line_pre_drop": "time_rel_line"}, inplace=True
            )

        elif not self.bidir and not self.keep_unidir:
            self.df_photons = self.df_photons.iloc[uneven_lines != 1, :].copy()
            self.df_photons.rename(
                columns={"time_rel_line_pre_drop": "time_rel_line"}, inplace=True
            )
            self.dict_of_data["Lines"] = (
                self.dict_of_data["Lines"].iloc[::2, :].copy().reset_index()
            )

        elif (
            not self.bidir and self.keep_unidir
        ):  # Unify the excess rows and photons in them into the previous row
            self.sorted_indices[np.logical_and(uneven_lines, 1)] -= 1
            self.df_photons.loc["Lines"] = (
                self.dict_of_data["Lines"].loc[self.sorted_indices, "abs_time"].values
            )
            self.dict_of_data["Lines"] = (
                self.dict_of_data["Lines"].iloc[::2, :].copy().reset_index()
            )
        try:
            self.df_photons.drop(["time_rel_line_pre_drop"], axis=1, inplace=True)
        except (ValueError, KeyError):  # column label doesn't exist
            pass
        self.df_photons = self.df_photons.loc[
            self.df_photons.loc[:, "time_rel_line"] >= 0, :
        ]

    def __reindex_dict_of_data(self):
        """
        Add new frame indices to the Series composing ``self.dict_of_data`` for slicing later on.
        The "Frames" indices are its values, and the "Lines" indices are the corresponding frames.
        """
        # Frames
        self.dict_of_data["Frames"] = pd.Series(
            self.dict_of_data["Frames"].abs_time.to_numpy(),
            index=self.dict_of_data["Frames"].abs_time.to_numpy(),
        )
        # Lines
        lines = self.dict_of_data["Lines"].abs_time.to_numpy()
        sorted_indices = np.digitize(lines, self.dict_of_data["Frames"].values) - 1
        positive_mask = sorted_indices >= 0
        lines = lines[positive_mask].copy()
        sorted_indices = sorted_indices[positive_mask]
        self.dict_of_data["Lines"] = pd.Series(
            lines, index=self.dict_of_data["Frames"].iloc[sorted_indices].values
        )

    def _rotate_laser_timings(self):
        """Rotate the histogram of the photon arrival times relative to the laser pulse,
        so that the highest value is in the second bin."""
        # bins_for_flim_hist = self._find_integer_gcd() + 1
        # hist, _ = np.histogram(self.df_photons["time_rel_pulse"], np.arange(bins_for_flim_hist))
        # peaks, params = find_peaks(hist, height=(None, None))
        # if len(peaks) == 0:
        #     return
        # max_peak_idx = (peaks[np.argmax(params["peak_heights"])] - 2) % bins_for_flim_hist
        # self.df_photons["time_rel_pules"] = (self.df_photons["time_rel_pulse"] - max_peak_idx) % bins_for_flim_hist
        pass



import pytest
import pandas as pd
import numpy as np

from pysight.nd_hist_generator.allocation import *


class TestAllocation:

    df_photons = pd.DataFrame(np.arange(10), columns=["abs_time"])
    dict_of_data = dict(
        Frames=pd.DataFrame(
            np.arange(0, 100, 10, dtype=np.uint64), columns=["abs_time"]
        ),
        Lines=pd.DataFrame(np.arange(0, 100, 5, dtype=np.uint64), columns=["abs_time"]),
    )
    # allocat = Allocate(df_photons.copy(), dict_of_data.copy())

    def test_reindex_dict_of_data_standard_frames(self):
        allocat = Allocate(self.df_photons.copy(), self.dict_of_data.copy())
        allocat._Allocate__reindex_dict_of_data()
        true_frames = pd.Series(
            data=self.dict_of_data["Frames"].abs_time.values,
            index=self.dict_of_data["Frames"].abs_time.values,
        )
        np.testing.assert_equal(
            allocat.dict_of_data["Frames"].values, true_frames.values
        )
        np.testing.assert_equal(
            allocat.dict_of_data["Frames"].index.values, true_frames.index.values
        )

    def test_reindex_dict_of_data_standard_lines(self):
        allocat = Allocate(self.df_photons.copy(), self.dict_of_data.copy())
        allocat._Allocate__reindex_dict_of_data()
        lines = self.dict_of_data["Lines"].abs_time.values
        frames = np.repeat(self.dict_of_data["Frames"].abs_time.values.copy(), 2)
        np.testing.assert_equal(allocat.dict_of_data["Lines"].values, lines)
        np.testing.assert_equal(allocat.dict_of_data["Lines"].index.values, frames)

    def test_reindex_dict_of_data_lines_before_start_of_frames(self):
        frames = pd.DataFrame(
            np.arange(30, 100, 10, dtype=np.uint64), columns=["abs_time"]
        )
        lines = pd.DataFrame(
            np.arange(0, 100, 5, dtype=np.uint64), columns=["abs_time"]
        )
        dict_of_data = dict(Frames=frames, Lines=lines)
        allocat = Allocate(self.df_photons.copy(), dict_of_data)
        allocat._Allocate__reindex_dict_of_data()
        lines = lines.abs_time.values
        np.testing.assert_equal(allocat.dict_of_data["Lines"].values, lines[6:])
        np.testing.assert_equal(
            allocat.dict_of_data["Lines"].index.values,
            np.repeat(frames.abs_time.values.copy(), 2),
        )

    def test_reindex_dict_of_data_lines_much_after_last_frame(self):
        frames = pd.DataFrame(
            np.arange(30, 100, 10, dtype=np.uint64), columns=["abs_time"]
        )
        lines = pd.DataFrame(
            np.arange(0, 150, 5, dtype=np.uint64), columns=["abs_time"]
        )
        dict_of_data = dict(Frames=frames, Lines=lines)
        allocat = Allocate(self.df_photons.copy(), dict_of_data)
        allocat._Allocate__reindex_dict_of_data()
        lines = lines.abs_time.values
        np.testing.assert_equal(allocat.dict_of_data["Lines"].values, lines[6:])
        frames = np.repeat(frames.abs_time.values.copy(), 2)
        frames = np.concatenate((frames, np.repeat(np.array([90]), 10)))
        np.testing.assert_equal(allocat.dict_of_data["Lines"].index.values, frames)

    @pytest.mark.parametrize(
        "binwidth,laser_freq,expected",
        [
            (800e-12, 80e6, 125),
            (800e-12, 80.3e6, 125),
            (400e-12, 80e6, 125),
            (200e-12, 80.3e6, 125),
            (100e-12, 80.4e6, 125),
        ],
    )
    def test_gcd(self, binwidth, laser_freq, expected):
        frames = pd.DataFrame(
            np.arange(30, 100, 10, dtype=np.uint64), columns=["abs_time"]
        )
        lines = pd.DataFrame(
            np.arange(0, 150, 5, dtype=np.uint64), columns=["abs_time"]
        )
        dict_of_data = dict(Frames=frames, Lines=lines)
        allocat = Allocate(
            self.df_photons.copy(),
            dict_of_data,
            binwidth=binwidth,
            laser_freq=laser_freq,
        )
        gcd = allocat._find_integer_gcd()
        assert gcd == expected

    @pytest.mark.parametrize("binwidth,laser_freq,expected", [(800e-12, 80e6, 8)])
    def test_num_pulse(self, binwidth, laser_freq, expected):
        frames = pd.DataFrame(
            np.arange(30, 100, 10, dtype=np.uint64), columns=["abs_time"]
        )
        lines = pd.DataFrame(
            np.arange(0, 150, 5, dtype=np.uint64), columns=["abs_time"]
        )
        dict_of_data = dict(Frames=frames, Lines=lines)
        allocat = Allocate(
            self.df_photons.copy(),
            dict_of_data,
            binwidth=binwidth,
            laser_freq=laser_freq,
        )
        needed_bins = allocat._find_integer_gcd()
        num_pulses = allocat._find_num_pulses(needed_bins)
        assert num_pulses == expected



"""
__author__ = Hagai Hargil
"""
import attr
import numpy as np
import pandas as pd
from attr.validators import instance_of
from pysight.nd_hist_generator.movie import Movie, FrameChunk
from collections import deque, namedtuple
from typing import Tuple, Union
from numba import jit, uint8, int64


@attr.s(slots=True)
class CensorCorrection(object):
    raw = attr.ib(validator=instance_of(dict))
    data = attr.ib(validator=instance_of(pd.DataFrame))
    movie = attr.ib(validator=instance_of(Movie))
    all_laser_pulses = attr.ib()
    nano_flim_list = attr.ib(init=False)
    flim = attr.ib(default=False, validator=instance_of(bool))
    reprate = attr.ib(default=80e6, validator=instance_of(float))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    laser_offset = attr.ib(default=3.5, validator=instance_of(float))
    num_of_channels = attr.ib(default=1, validator=instance_of(int))

    @property
    def bins_bet_pulses(self) -> int:
        return int(np.ceil(1 / (self.reprate * self.binwidth)))

    @property
    def offset(self):
        return int(np.floor(self.laser_offset * 10 ** -9 / self.binwidth))

    def run(self):
        """
        Main pipeline for the censor correction part.
        """
        if self.flim:
            print("Starting the censor correction...")
            self.create_arr_of_hists_deque()
        else:
            print("FLIM deactivated, no censor correction performed.")

    def __gen_laser_pulses_deque(self) -> np.ndarray:
        """
        If data has laser pulses - return them. Else - simulate them with an offset
        """
        start_time = 0
        step = self.bins_bet_pulses
        volumes_in_movie = self.movie.gen_of_volumes()

        if (
            self.all_laser_pulses == 0 and self.flim == False
        ):  # no 'Laser' data was recorded
            for vol in volumes_in_movie:
                yield np.arange(
                    start=start_time + self.offset,
                    stop=vol.end_time,
                    step=step,
                    dtype=np.uint64,
                )

        elif self.all_laser_pulses == 0 and self.flim == True:
            pass

        else:
            for vol in volumes_in_movie:
                yield self.all_laser_pulses[
                    (self.all_laser_pulses >= vol.abs_start_time - step)
                    & (self.all_laser_pulses <= vol.end_time + step)
                ] + self.offset

    def __get_bincount_deque(self):
        print("Movie object created. Generating the bincount deque...")
        bincount_deque = deque()
        laser_pulses_deque = self.__gen_laser_pulses_deque()
        volumes_in_movie = self.movie.gen_of_volumes()
        for idx, vol in enumerate(volumes_in_movie):
            censored = CensoredVolume(
                df=vol.data,
                vol=vol,
                offset=self.offset,
                laser_pulses=next(laser_pulses_deque),
            )
            dig, bincount = censored.gen_bincount()
            pos_idx = np.where(dig >= 0)[0]
            dig = dig[pos_idx]
            pos_photons = censored.df.iloc[pos_idx, -1].values.T
            if len(pos_photons) == 0:
                data_dict = {
                    "photon_hist": np.zeros((self.bins_bet_pulses, 1), dtype=np.uint8),
                    "bincount": bincount,
                    "num_empty_hists": bincount[0],
                }
                return data_dict

            photon_hist = np.zeros(
                (self.bins_bet_pulses, pos_photons.shape[0]), dtype=np.uint8
            )
            for laser_idx, photon in enumerate(np.nditer(pos_photons)):
                start_time = censored.laser_pulses[dig[laser_idx]]
                try:
                    end_time = censored.laser_pulses[dig[laser_idx] + 1]
                except IndexError:  # photons out of laser pulses
                    continue
                else:
                    photon_hist[:, laser_idx] = np.histogram(
                        photon, bins=np.arange(start_time, end_time + 1, dtype="uint64")
                    )[0]

            data_dict = {
                "photon_hist": photon_hist,
                "bincount": bincount,
                "num_empty_hists": bincount[0],
            }
            assert (
                data_dict["num_empty_hists"] >= 0
            ), "Sum of bincount: {}, number of photons: {}".format(
                sum(bincount), laser_idx
            )
            bincount_deque.append(data_dict)

        return bincount_deque

    def find_temporal_structure_deque(self):
        temp_struct_deque = deque()
        laser_pulses_deque = self.__gen_laser_pulses_deque()
        volumes_in_movie = self.movie.gen_of_volumes()
        for idx, vol in enumerate(volumes_in_movie):
            censored = CensoredVolume(
                df=vol.data,
                vol=vol,
                offset=self.offset,
                laser_pulses=next(laser_pulses_deque),
                binwidth=self.binwidth,
                reprate=self.reprate,
            )
            temp_struct_deque.append(censored.find_temp_structure())
        return temp_struct_deque

    def __worker_arr_of_hists(self, vol):
        censored = CensoredVolume(
            df=vol.data,
            vol=vol,
            offset=self.offset,
            binwidth=self.binwidth,
            reprate=self.reprate,
        )
        return censored.gen_arr_of_hists()

    def create_arr_of_hists_deque(self):
        """
        For each volume generate a single matrix with the same size as the underlying volume,
        which contains a histogram of photons in their laser pulses for each pixel.
        :return: deque() that contains an array of histograms in each place
        """
        self.nano_flim_list = []  # each cell contains a different data channel
        for chan in range(1, self.num_of_channels + 1):
            print("Starting channel number {}: ".format(chan))
            volumes_in_movie = self.movie.gen_of_volumes(channel_num=chan)
            self.nano_flim_list.append(
                [self.__worker_arr_of_hists(vol) for vol in volumes_in_movie]
            )

    def create_array_of_hists_deque(self):
        """
        Go through each volume in the deque and find the laser pulses for each pixel, creating a summed histogram per pixel.
        :return:
        """
        temp_struct_deque = deque()
        laser_pulses_deque = self.__gen_laser_pulses_deque()
        volumes_in_movie = self.movie.gen_of_volumes()
        for idx, vol in enumerate(volumes_in_movie):
            censored = CensoredVolume(
                df=vol.data,
                vol=vol,
                offset=self.offset,
                laser_pulses=next(laser_pulses_deque),
                binwidth=self.binwidth,
                reprate=self.reprate,
            )
            temp_struct_deque.append(censored.gen_array_of_hists())
        return temp_struct_deque

    def __gen_labels(self, size: int, label: Union[int, float]) -> np.ndarray:
        """
        Create labels for the ML algorithm. Label value must be an integer.
        :size: Number of elements
        :return: np.ndarray
        """
        if isinstance(label, int):  # fixed power during the session
            return np.ones(size, dtype=np.uint8) * label
        elif isinstance(
            label, float
        ):  # `label` contains the frequency of the triangular wave
            pass

    def learn_histograms(
        self, label: Union[int, float], power: int, folder_to_save: str
    ):
        """
        Implement the machine learning algorithm on the data.
        :param label: Label of ML algorithm.
        :param power: How much power was injected to the Qubig. For saving the file.
        :return: data, labels
        """
        from sklearn import svm, metrics
        import pathlib

        # Start by generating the data and arranging it properly for the machine
        bincount = self.__get_bincount_deque()
        print("Bincount done. Adding all data to a single matrix.")

        data = np.empty((self.bins_bet_pulses, 0))
        for vol in bincount:
            data = np.concatenate(
                (data, vol["photon_hist"]), axis=1
            )  # the histograms with photons in them
            data = np.concatenate(
                (
                    data,
                    np.zeros(
                        (self.bins_bet_pulses, vol["num_empty_hists"]), dtype=np.uint8
                    ),
                ),
                axis=1,
            )  # empty hists
        data = data.T
        n_samples = data.shape[0]
        labels = self.__gen_labels(n_samples, label)
        classifier = svm.SVC(gamma=0.001)
        labels[1] = 10  # toying around

        print("Fitting the data...")
        classifier.fit(data[: n_samples // 2], labels[: n_samples // 2])

        # Predictions
        expected = labels[n_samples // 2 :]
        predicted = classifier.predict(data[n_samples // 2 :])
        print("Number of samples is %s." % n_samples)
        print(
            "Classification report for classifier %s:\n%s\n"
            % (classifier, metrics.classification_report(expected, predicted))
        )
        print("Confusion matrix:\n%s" % metrics.confusion_matrix(expected, predicted))

        # Save the data for future use
        folder_as_path = pathlib.Path(folder_to_save)
        filename = str(folder_as_path / "{}p_label_{}.npy".format(power, label))
        self.__save_data(data=data, filename=filename)
        return data, labels

    def __save_data(self, data: np.ndarray, filename: str):
        """
        Save the data array for future training.
        :param data: Data to be learnt.
        :param filename: Including dir
        :return:
        """
        print("Saving to {}...".format(filename))
        with open(filename, "wb") as f:
            np.save(f, data)

    def append_laser_line(self):
        """
        Add a final laser line to the laser signal input.
        """
        last_laser_row = pd.DataFrame(
            {
                "abs_time": self.raw["Laser"]["abs_time"].iat[-1]
                + self.bins_bet_pulses,
                "edge": 0,
                "sweep": self.raw["Laser"]["sweep"].iat[-1],
                "time_rel_sweep": self.raw["Laser"]["time_rel_sweep"].iat[-1]
                + self.bins_bet_pulses,
            },
            index=[self.raw["Laser"].shape[0]],
        )
        self.raw["Laser"] = pd.concat([self.raw["Laser"], last_laser_row])

    def train_dataset(self):
        """
        Using almost raw data, allocate photons to their laser pulses
        (instead of laser pulses to photons) and create all 16 bit words for the ML algorithm.
        :return:
        """
        # Append a fake laser pulse to retain original number of "bins"
        self.append_laser_line()
        sorted_indices = pd.cut(
            self.raw["PMT1"]["abs_time"],
            bins=self.raw["Laser"]["abs_time"],
            labels=self.raw["Laser"].iloc[:-1, 3],
            include_lowest=True,
        )
        self.raw["Laser"].set_index(
            keys="time_rel_sweep", inplace=True, append=True, drop=True
        )
        num_of_pos_bins = 22
        new_bins = np.arange(-10, num_of_pos_bins + 1)  # 32 bins
        min_time_after_sweep = 10
        max_time_after_sweep = (
            self.raw["Laser"]["time_rel_sweep"].max() - num_of_pos_bins - 1
        )
        indices = np.arange(min_time_after_sweep, max_time_after_sweep)
        hist_df = pd.DataFrame([], dtype=object)
        for idx in indices:
            cur_pulse = self.raw["Laser"].xs(
                idx, level="time_rel_sweep", drop_level=False
            )


@attr.s(slots=True)
class CensoredVolume(object):
    df = attr.ib(validator=instance_of(pd.DataFrame))
    chunk = attr.ib(validator=instance_of(FrameChunk))
    offset = attr.ib(validator=instance_of(int))
    binwidth = attr.ib(default=800e-12)
    reprate = attr.ib(default=80e6)

    @property
    def bins_bet_pulses(self) -> int:
        return int(np.ceil(1 / (self.reprate * self.binwidth)))

    def gen_bincount(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Bin the photons into their relative laser pulses, and count how many photons arrived due to each pulse.
        """
        hist, _ = np.histogram(
            self.chunk.data.loc[:, "time_rel_frames"].values, bins=self.laser_pulses
        )
        dig = (
            np.digitize(
                self.chunk.data.loc[:, "time_rel_frames"].values, bins=self.laser_pulses
            )
            - 1
        )
        return dig, np.bincount(hist)

    def find_temp_structure(self) -> np.ndarray:
        """
        Generate a summed histogram of the temporal structure of detected photons.
        """
        bins = np.arange(
            start=0, stop=np.ceil(1 / (self.reprate * self.binwidth)) + 1, step=1
        )
        subtracted_times, _, _ = self.sort_photons_in_pulses()
        hist, _ = np.histogram(subtracted_times, bins=bins)
        return hist

    def sort_photons_in_pulses(self):
        """
        Helper function to generate a searchsorted output of photons in laser pulses.
        """
        pulses = self.laser_pulses
        sorted_indices: np.ndarray = np.searchsorted(
            pulses, self.movie.data["time_rel_frames"].values
        ) - 1
        array_of_laser_starts = pulses[sorted_indices]
        subtracted_times = (
            self.movie.data["time_rel_frames"].values - array_of_laser_starts
        )
        return subtracted_times, array_of_laser_starts, sorted_indices

    def gen_array_of_hists(self) -> np.ndarray:
        """
        For a specific frame, sort photons and laser pulses inside the pixels to gain
        statistics on the distribution of the photons inside the pixels.
        :return: np.ndarray of the same size as the original image. Each pixels contains
        a histogram inside it.
        """
        BinData = namedtuple("BinData", ("hist", "pulses", "photons"))
        all_pulses = 0
        all_photons = 0

        hist, edges = self.chunk.create_hist()
        # Create a relative timestamp to the line signal for each laser pulse
        sorted_pulses = np.searchsorted(edges[0][:-1], self.laser_pulses) - 1
        pulses = pd.DataFrame(
            data=self.laser_pulses[np.where(sorted_pulses >= 0)[0]],
            columns=["time_rel_frames"],
        )
        pulses = pulses.assign(
            Lines=edges[0][:-1][sorted_pulses[np.where(sorted_pulses >= 0)[0]]]
        )
        pulses.dropna(how="any", inplace=True)
        pulses.loc[:, "Lines"] = pulses.loc[:, "Lines"].astype("uint64")
        pulses.loc[:, "time_rel_line"] = (
            pulses.loc[:, "time_rel_frames"] - pulses.loc[:, "Lines"]
        )
        pulses.loc[:, "Lines"] = pulses.loc[:, "Lines"].astype("category")
        pulses.set_index(keys=["Lines"], inplace=True, append=True, drop=True)

        # Allocate laser pulses and photons to their bins
        pulses.loc[:, "bins_x"] = (
            np.digitize(pulses.loc[:, "time_rel_frames"].values, bins=edges[0]) - 1
        ).astype("uint16", copy=False)
        pulses.loc[:, "bins_y"] = (
            np.digitize(pulses.loc[:, "time_rel_line"].values, bins=edges[1]) - 1
        ).astype("uint16", copy=False)
        self.chunk.data.loc[:, "bins_x"] = (
            np.digitize(self.chunk.data.loc[:, "time_rel_frames"].values, bins=edges[0])
            - 1
        ).astype("uint16", copy=False)
        self.chunk.data.loc[:, "bins_y"] = (
            np.digitize(self.chunk.data.loc[:, "time_rel_line"].values, bins=edges[1])
            - 1
        ).astype("uint16", copy=False)
        pulses.set_index(
            keys=["bins_x", "bins_y"], inplace=True, append=True, drop=True
        )
        self.chunk.data.set_index(
            keys=["bins_x", "bins_y"], inplace=True, append=True, drop=True
        )

        # Go through each bin and histogram the photons there
        image_bincount = np.zeros_like(hist, dtype=object)
        for row in range(self.chunk.x_pixels):
            row_pulses = pulses.xs(key=row, level="bins_x", drop_level=False)
            assert len(row_pulses) > 0, "Row {} contains no pulses".format(row)
            try:
                row_photons = self.chunk.data.xs(
                    key=row, level="bins_x", drop_level=False
                )
            except KeyError:  # no photons in row
                for col in range(self.chunk.y_pixels):
                    final_pulses = row_pulses.xs(
                        key=col, level="bins_y", drop_level=False
                    )
                    hist = (
                        np.histogram(
                            np.array([]),
                            bins=final_pulses.loc[:, "time_rel_line"].values,
                        )[0]
                    ).astype("uint8")
                    cur_bincount = np.bincount(hist).astype("uint64", copy=False)
                    tot_pulses = np.sum(cur_bincount)
                    tot_photons = np.average(
                        cur_bincount, weights=range(len(cur_bincount))
                    )
                    image_bincount[row, col] = BinData(
                        hist=hist, pulses=tot_pulses, photons=tot_photons
                    )
                    all_photons += tot_photons
                    all_pulses += tot_pulses
            else:
                for col in range(self.movie.y_pixels):
                    final_pulses = row_pulses.xs(
                        key=col, level="bins_y", drop_level=False
                    )
                    assert (
                        len(final_pulses) > 0
                    ), "Column {} in row {} contains no pulses".format(col, row)
                    try:
                        final_photons = row_photons.xs(
                            key=col, level="bins_y", drop_level=False
                        )
                    except KeyError:  # no photons in col
                        hist = (
                            np.histogram(
                                np.array([]),
                                bins=final_pulses.loc[:, "time_rel_line"].values,
                            )[0]
                        ).astype("uint8", copy=False)
                    else:
                        # hist = (np.histogram(final_photons.loc[:, 'time_rel_line'].values,
                        #                     bins=final_pulses.loc[:, 'time_rel_line'].values)[0])\
                        #     .astype('uint8', copy=False)
                        hist = numba_histogram(
                            final_photons.loc[:, "time_rel_line"].values,
                            bins=final_pulses.loc[:, "time_rel_line"].values,
                        ).astype("uint8", copy=False)
                    finally:
                        if not np.all(hist >= 0):
                            print("WHAT IS GOING ON")
                        assert np.all(
                            hist >= 0
                        ), "In row {}, column {}, the histogram turned out to be negative.".format(
                            row, col
                        )
                        cur_bincount = numba_bincount(hist).astype("uint64", copy=False)
                        tot_pulses = np.sum(cur_bincount)
                        tot_photons = np.average(
                            cur_bincount, weights=range(len(cur_bincount))
                        )
                        image_bincount[row, col] = BinData(
                            hist=hist, pulses=tot_pulses, photons=tot_photons
                        )
                        all_photons += tot_photons
                        all_pulses += tot_pulses

        return image_bincount, all_pulses, all_photons

    def gen_arr_of_hists(self) -> np.ndarray:
        """
        For each specific frame, sort the photons inside the pixels to gqin
        statistics on the distribution of the photons inside the pixels.
        :return: np.ndarray of the same size as the original frame,
        with each bin containing a histogram.
        """
        edges = self.chunk._Volume__create_hist_edges()[0]

        hist_t, edges_t = self.chunk.create_hist()
        assert (
            "time_rel_pulse" in self.chunk.data.columns
        ), "No `time_rel_pulse` column in data."

        self.chunk.data.loc[:, "bins_x"] = (
            np.digitize(self.chunk.data.loc[:, "time_rel_frames"].values, bins=edges[0])
            - 1
        ).astype("uint16", copy=False)
        self.chunk.data.loc[:, "bins_y"] = (
            np.digitize(self.chunk.data.loc[:, "time_rel_line"].values, bins=edges[1])
            - 1
        ).astype("uint16", copy=False)
        self.chunk.data.set_index(
            keys=["bins_x", "bins_y"], inplace=True, append=True, drop=True
        )
        self.chunk.data.drop(len(edges[0]) - 1, level="bins_x", inplace=True)
        self.chunk.data.drop(len(edges[1]) - 1, level="bins_y", inplace=True)

        image_bincount = np.zeros(
            (len(edges[0]) - 1, len(edges[1]) - 1), dtype=object
        )  # returned variable, contains hists
        active_lines = np.unique(self.chunk.data.index.get_level_values("bins_x"))
        for row in active_lines:
            print("Row number: {}".format(row))
            row_photons = self.chunk.data.xs(key=row, level="bins_x", drop_level=False)
            rel_idx = np.unique(row_photons.index.get_level_values("bins_y"))
            image_bincount[row, rel_idx] = self.__allocate_photons_to_bins(
                idx=rel_idx, photons=row_photons
            )

        return image_bincount

    def __allocate_photons_to_bins(self, idx: np.ndarray, photons: pd.DataFrame):
        """
        Generate a summed-histogram of photons for the relevant edges.
        :param photons:
        :param edge_x:
        :param edge_y:
        :return:
        """
        hist_storage = np.zeros_like(idx, dtype=object)
        for cur_idx, col in enumerate(idx):
            cur_photons = photons.xs(key=col, level="bins_y", drop_level=False)
            hist_storage[cur_idx] = numba_histogram(
                cur_photons.loc[:, "time_rel_pulse"].values,
                bins=np.arange(0, self.bins_bet_pulses + 1, dtype=np.uint8),
            ).astype("uint8", copy=False)
        return hist_storage


@jit((int64[:](uint8[:], uint8[:])), nopython=True, cache=True)
def numba_histogram(arr: np.array, bins) -> np.array:
    return np.histogram(arr, bins)[0]


@jit((int64[:](uint8[:])), nopython=True, cache=True)
def numba_bincount(arr: np.array) -> np.array:
    return np.bincount(arr)


#
# @jit((int64[:](uint64[:], uint64[:])), nopython=True, cache=True)
# def numba_digitize(arr: np.array) -> np.array:
#     pass



import pandas as pd
import attr
from typing import Dict
from attr.validators import instance_of


@attr.s(slots=True)
class DistributeData:
    """
    Separates the channel-specific data to their own channels.

    :param pd.DataFrame df: Raw data
    :param dict dict_of_inputs: Mapping of inputs to data they contain
    :param bool use_tag_bits: Whether TAG bits are needed
    """

    df = attr.ib(validator=instance_of(pd.DataFrame))
    dict_of_inputs = attr.ib(validator=instance_of(dict))
    use_tag_bits = attr.ib(default=False, validator=instance_of(bool))
    dict_of_data = attr.ib(init=False)
    data_to_grab = attr.ib(init=False)

    def run(self) -> None:
        """ Runs the allocation function, populating self.dict_of_data """
        self.dict_of_data = self.__allocate_data_by_channel()

    def __allocate_data_by_channel(self) -> Dict:
        """
        Go over the channels and find the events from that specific channel, assigning
        them to a dictionary with a suitable name.

        :return dict: Dict containing the data
        """
        dict_of_data = {}
        self.data_to_grab = [
            "abs_time",
            "sweep",
        ]  # relevant columns of the DF for analysis
        if self.use_tag_bits:
            self.data_to_grab.extend(["tag", "edge"])
        for key in self.dict_of_inputs:
            relevant_values = self.df.loc[
                self.df["channel"] == self.dict_of_inputs[key], self.data_to_grab
            ]
            if key in ["PMT1", "PMT2"]:
                dict_of_data[key] = relevant_values.reset_index(drop=True)
                dict_of_data[key]["Channel"] = (
                    1 if "PMT1" == key else 2
                )  # channel is the spectral channel
            else:
                dict_of_data[key] = relevant_values.sort_values(
                    by=["abs_time"]
                ).reset_index(drop=True)

        return dict_of_data



import numpy as np
import pandas as pd
import pytest

from pysight.nd_hist_generator.line_signal_validators.add_bidir_lines import (
    add_bidir_lines,
)


@pytest.fixture
def even_data():
    lines = pd.DataFrame({"abs_time": np.arange(0, 11, 2)})
    data = dict(Lines=lines)
    return data


@pytest.fixture
def odd_data():
    lines = pd.DataFrame({"abs_time": np.arange(1, 12, 2)})
    data = dict(Lines=lines)
    return data


def test_add_even_lines(even_data):
    new_data = add_bidir_lines(even_data)
    correct = np.arange(0, 11, dtype=np.uint64)
    assert np.array_equal(new_data["Lines"].loc[:, "abs_time"].to_numpy(), correct)


def test_add_uneven_lines(odd_data):
    new_data = add_bidir_lines(odd_data)
    correct = np.arange(1, 12, dtype=np.uint64)
    assert np.array_equal(new_data["Lines"].loc[:, "abs_time"].to_numpy(), correct)



"""
__author__ = Hagai Har-Gil
"""
from unittest import TestCase
import numpy as np
import pandas as pd
from typing import Tuple

from pysight.nd_hist_generator.volume_gen import VolumeGenerator


def gen_test_df(frame_num=10, end=1_000_000) -> Tuple[pd.DataFrame, pd.Series]:
    photons = np.arange(0, end, dtype=np.uint64)
    frames = np.linspace(0, end, num=frame_num, dtype=np.uint64, endpoint=False)
    ones_frames = np.ones((1, int(len(photons) / len(frames))), dtype=np.uint64)
    frames_ser = pd.Series(frames)
    frames = (np.atleast_2d(frames).T @ ones_frames).ravel()
    assert len(frames) == len(photons)
    df = pd.DataFrame({"time_rel_frame": photons - frames, "Frames": frames})
    df.set_index(["Frames"], drop=True, inplace=True)
    return df, frames_ser


class TestVolumeGenerator(TestCase):
    def test_many_frames(self):
        _, frames = gen_test_df(1000)
        shape = (1000, 512, 512, 16)
        max_vols = int(300e6 / (np.prod(shape[1:]) * 8))
        volgen = VolumeGenerator(frames, shape, MAX_BYTES_ALLOWED=int(300e6))
        volgen.create_frame_slices(create_slices=False)
        chunks = list(volgen.full_frame_chunks)
        self.assertEqual(len(chunks[0]), max_vols)

    def test_single_frame(self):
        _, frames = gen_test_df(1000)
        shape = (10, 512, 512, 16)
        max_vols = 1
        volgen = VolumeGenerator(frames, shape, MAX_BYTES_ALLOWED=100)
        volgen.create_frame_slices(create_slices=False)
        self.assertEqual(len(list(volgen.full_frame_chunks)[0]), max_vols)

    def test_standard_slice(self):
        _, frames = gen_test_df(10, end=100)
        shape = (10, 512, 512, 16)
        volgen = VolumeGenerator(frames, shape, MAX_BYTES_ALLOWED=int(300e6))
        vol_times = volgen.create_frame_slices()
        self.assertSequenceEqual(
            list(vol_times), [slice(0, 70, None), slice(80, 90, None)]
        )

    def test_full_slice(self):
        _, frames = gen_test_df(16, end=1600)
        shape = (16, 512, 512, 16)
        volgen = VolumeGenerator(frames, shape, MAX_BYTES_ALLOWED=int(300e6))
        vol_times = volgen.create_frame_slices()
        self.assertSequenceEqual(list(vol_times), [slice(0, 700), slice(800, 1500)])

    def test_single_slice(self):
        _, frames = gen_test_df(1, end=10)
        shape = (1, 512, 512, 16)
        volgen = VolumeGenerator(frames, shape)
        vol_times = volgen.create_frame_slices()
        self.assertSequenceEqual(list(vol_times), [slice(0, 0)])

    def test_single_chunk(self):
        _, frames = gen_test_df(10, end=100)
        shape = (10, 800, 800, 100)
        volgen = VolumeGenerator(frames, shape, MAX_BYTES_ALLOWED=int(300e6))
        vol_times = volgen.create_frame_slices()
        self.assertSequenceEqual(
            list(vol_times),
            [
                slice(0, 0, None),
                slice(10, 10, None),
                slice(20, 20, None),
                slice(30, 30, None),
                slice(40, 40, None),
                slice(50, 50, None),
                slice(60, 60, None),
                slice(70, 70, None),
                slice(80, 80, None),
                slice(90, 90, None),
            ],
        )

    def test_grouper(self):
        volgen = VolumeGenerator(pd.Series(), (1,))
        volgen.frames = pd.Series([10, 20, 30, 40], dtype=np.uint64)
        volgen.frames_per_chunk = 3
        grouped = volgen._VolumeGenerator__grouper()
        self.assertSequenceEqual(list(grouped), [(10, 20, 30), (40, np.nan, np.nan)])

    def test_grouper_single_chunk(self):
        volgen = VolumeGenerator(pd.Series(), (1,))
        volgen.frames = pd.Series([10, 20, 30, 40], dtype=np.uint64)
        volgen.frames_per_chunk = 1
        grouped = volgen._VolumeGenerator__grouper()
        self.assertSequenceEqual(list(grouped), [(10,), (20,), (30,), (40,)])



from unittest import TestCase
from pysight.nd_hist_generator.movie import *
from pysight.nd_hist_generator.volume_gen import *
import pandas as pd
import numpy as np


def gen_data_df(frame_num=10, line_num=1000, end=100_000):
    """
    Mock data for tests.
    Returns:
        df - The full DataFrame
        frames only
        lines only
        x pixels
        y pixels
    """
    photons = np.arange(0, end, dtype=np.uint64)
    channel = np.ones_like(photons)
    lines = np.linspace(0, end, num=line_num, endpoint=False, dtype=np.uint64)
    x_pix = int(len(photons) / len(lines))
    ones_lines = np.ones((1, int(len(photons) / len(lines))), dtype=np.uint64)
    frames = np.linspace(0, end, num=frame_num, dtype=np.uint64, endpoint=False)
    frames_ser = pd.Series(frames)
    ones_frames = np.ones((1, int(len(photons) / len(frames))), dtype=np.uint64)
    lines = (np.atleast_2d(lines).T @ ones_lines).ravel()
    frames = (np.atleast_2d(frames).T @ ones_frames).ravel()
    assert len(lines) == len(frames) == len(photons)

    df = pd.DataFrame(
        {
            "abs_time": photons,
            "time_rel_line": photons - lines,
            "Lines": lines,
            "Frames": frames,
            "Channel": channel,
        }
    )
    df["Channel"] = df["Channel"].astype("category")
    df.set_index(["Channel", "Frames", "Lines"], drop=True, inplace=True)
    y_pix = x_pix
    return df, frames_ser, pd.Series(np.unique(lines)), x_pix, y_pix


class TestMovies(TestCase):
    frame_num = 10
    end = 1000
    line_num = 100
    data, frames, lines, x_pix, y_pix = gen_data_df(
        frame_num=frame_num, line_num=line_num, end=end
    )
    data_shape = (frame_num, x_pix, y_pix)
    volgen = VolumeGenerator(frames, data_shape)
    fr = volgen.create_frame_slices()
    movie = Movie(
        data=data,
        lines=lines,
        data_shape=data_shape,
        outputs={"memory": True},
        line_delta=int(lines.diff().mean()),
        fill_frac=100.0,
        bidir=True,
        frame_slices=fr,
        frames=frames,
        frames_per_chunk=volgen.frames_per_chunk,
    )
    movie.run()

    def test_all_pipeline_basic(self):
        self.assertTrue(
            np.all(self.movie.stack[1].ravel() == np.ones((self.end,), dtype=np.uint8))
        )

    def test_baseline_outputs(self):
        during, end = self.movie._Movie__determine_outputs()
        self.assertTrue(len(during) == 1)
        self.assertTrue(len(end) == 1)
        during, end = str(during), str(end)
        self.assertTrue("create_memory_output" in during)
        self.assertTrue("convert_list_to_arr" in end)

    def test_slice_df(self):
        sl = slice(0, 23000)
        movie = Movie(
            self.data,
            self.lines,
            data_shape=self.data_shape,
            outputs={"memory": True},
            line_delta=int(self.lines.diff().mean()),
            fill_frac=100.0,
            bidir=True,
            frames=self.frames,
            frame_slices=(slice(1) for n in range(2)),
        )
        di = movie._Movie__slice_df(sl)
        self.assertTrue(1 in di[0].keys())
        self.assertSequenceEqual((di[0][1].shape, di[1]), ((1000, 2), 10))

    def test_single_slice_df(self):
        sl = slice(0, 0)
        movie = Movie(
            self.data,
            self.lines,
            data_shape=self.data_shape,
            outputs={"memory": True},
            line_delta=int(self.lines.diff().mean()),
            fill_frac=100.0,
            bidir=True,
            frames=self.frames,
            frame_slices=(slice(1) for n in range(2)),
        )
        di = movie._Movie__slice_df(sl)
        self.assertTrue(1 in di[0].keys())
        self.assertSequenceEqual((di[0][1].shape, di[1]), ((100, 2), 1))



import attr
from attr.validators import instance_of
import pandas as pd
import numpy as np

from pysight.ascii_list_file_parser.apply_df_funcs import convert_bin_to_int
from pysight.ascii_list_file_parser.tabulation import slice_string_arrays


@attr.s(slots=True)
class ParseTAGBits(object):
    """
    Take TAG bits and add their value to the photon dataframe
    """

    dict_of_data = attr.ib(validator=instance_of(dict))
    photons = attr.ib(validator=instance_of(pd.DataFrame))
    bits_dict = attr.ib(validator=instance_of(dict))
    use_tag_bits = attr.ib(default=False, validator=instance_of(bool))
    df_with_bits = attr.ib(init=False)

    @property
    def func_for_keyword(self):
        return {
            "Power": self.__parse_power,
            "Z axis": self.__parse_tag_lens,
            "Slow axis": self.__parse_slow_axis,
            "Fast axis": self.__parse_fast_axis,
        }

    def gen_df(self):
        """
        Add the TAG bits as a column
        :return: pd.DataFrame: A new instance of the photon dataframe
        """
        if not self.use_tag_bits:
            self.photons.drop("tag", axis=1, inplace=True, errors="ignore")
            return self.photons

        assert "tag" in self.dict_of_data["PMT1"].columns

        for key in self.bits_dict:
            self.func_for_keyword[key]()

        return self.photons

    def __parse_power(self):
        """
        Parse the bits of a power modulator
        :return:
        """
        proc_tag_bits = slice_string_arrays(
            self.dict_of_data["PMT1"].tag.values,
            start=self.bits_dict.start,
            end=self.bits_dict.end + 1,
        )

        int_tag_bits = convert_bin_to_int(proc_tag_bits)
        self.photons.loc[:, "Power Factor"] = int_tag_bits

    def __parse_tag_lens(self):
        pass

    def __parse_slow_axis(self):
        pass

    def __parse_fast_axis(self):
        pass
